import unittest

import itertools
from parameterized import parameterized

from tnsgrt.michell import Michell


class TestMichell(unittest.TestCase):

    @parameterized.expand(itertools.product([4, 6, 12], [3, 4, 10, 12], [0.5, 1, 1.5]))
    def test_michell(self, p, q, radius):
        s = Michell(p, q=q, radius=radius)
        # print(f'p = {p}, q = {q}')
        self.assertEqual(s.get_number_of_nodes(), p*q+1)
        # self.assertEqual(s.get_number_of_members(), p*(q+1)+2*p)
        # self.assertEqual(set(s.member_tags.keys()),
        # {'bar', 'string', 'vertical', 'top', 'bottom'})
        #
        # rho = 1
        # alpha = np.pi/2-np.pi/p
        # lambda_ = np.hstack((
        #     rho * np.cos(np.pi/p) / np.cos(alpha - np.pi/p) * np.ones((p,)),
        #     (1/rho) * np.cos(np.pi / p) / np.cos(alpha - np.pi/p) * np.ones((p,)),
        #     np.ones((p,)),
        #     -np.ones((p,))
        # ))
        # np.testing.assert_allclose(s.member_properties['lambda_'], lambda_)


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
import pandas as pd
import scipy

from tnsgrt import utils, structure
from tnsgrt.stiffness import NodeConstraint
from tnsgrt.structure import Structure


class TestStructure(unittest.TestCase):

    def test_constructor(self):

        # empty structure
        s = Structure()
        self.assertEqual(s.label, None)
        np.testing.assert_array_equal(s.nodes, np.zeros((3, 0), np.float_))
        self.assertEqual(len(s.node_properties), 0)
        self.assertEqual(s.get_number_of_nodes(), 0)
        self.assertEqual(s.node_tags, {})
        self.assertEqual(len(s.member_properties), 0)
        self.assertEqual(s.get_number_of_members(), 0)
        np.testing.assert_array_equal(s.members, np.zeros((2, 0), np.int64))
        np.testing.assert_array_equal(s.member_tags['bar'], np.zeros((0,), np.int64))
        np.testing.assert_array_equal(s.member_tags['string'], np.zeros((0,), np.int64))

        nodes = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        s = Structure(nodes, label='label')
        self.assertEqual(s.label, 'label')
        self.assertEqual(s.nodes.dtype, np.float_)
        self.assertEqual(s.members.dtype, np.int64)
        np.testing.assert_array_equal(s.nodes, nodes)
        self.assertEqual(len(s.node_properties), 3)
        self.assertEqual(s.get_number_of_nodes(), 3)
        self.assertEqual(s.node_tags, {})
        np.testing.assert_array_equal(s.members, np.zeros((2, 0), np.int64))
        np.testing.assert_array_equal(s.member_tags['bar'], np.zeros((0,), np.int64))
        np.testing.assert_array_equal(s.member_tags['string'], np.zeros((0,), np.int64))
        self.assertEqual(s.get_number_of_nodes(), 3)
        self.assertEqual(s.get_number_of_members(), 0)
        self.assertEqual(len(s.member_properties), 0)

        members = np.array([[0, 1, 2], [1, 2, 0]])
        s = Structure(nodes, members, label='label')
        self.assertEqual(s.label, 'label')
        self.assertEqual(s.nodes.dtype, np.float_)
        self.assertEqual(s.members.dtype, np.int64)
        np.testing.assert_array_equal(s.nodes, nodes)
        self.assertEqual(len(s.node_properties), 3)
        self.assertEqual(s.get_number_of_nodes(), 3)
        self.assertEqual(s.node_tags, {})
        np.testing.assert_array_equal(s.members, members)
        np.testing.assert_array_equal(s.member_tags['bar'],
                                      np.arange(0, 3, dtype=np.int64))
        np.testing.assert_array_equal(s.member_tags['string'], np.zeros((0,), np.int64))
        self.assertEqual(s.get_number_of_nodes(), 3)
        self.assertEqual(s.get_number_of_members(), 3)
        self.assertEqual(len(s.member_properties), 3)

        s = Structure(nodes, members, number_of_strings=1)
        self.assertEqual(s.label, None)
        self.assertEqual(s.nodes.dtype, np.float_)
        self.assertEqual(s.members.dtype, np.int64)
        np.testing.assert_array_equal(s.nodes, nodes)
        self.assertEqual(len(s.node_properties), 3)
        self.assertEqual(s.get_number_of_nodes(), 3)
        self.assertEqual(s.node_tags, {})
        np.testing.assert_array_equal(s.members, members)
        np.testing.assert_array_equal(s.member_tags['bar'],
                                      np.arange(1, 3, dtype=np.int64))
        np.testing.assert_array_equal(s.member_tags['string'],
                                      np.arange(0, 1, dtype=np.int64))
        self.assertEqual(s.get_number_of_nodes(), 3)
        self.assertEqual(s.get_number_of_members(), 3)
        self.assertEqual(len(s.member_properties), 3)

        s = Structure(nodes, members, number_of_strings=2)
        self.assertEqual(s.label, None)
        self.assertEqual(s.nodes.dtype, np.float_)
        self.assertEqual(s.members.dtype, np.int64)
        np.testing.assert_array_equal(s.nodes, nodes)
        self.assertEqual(len(s.node_properties), 3)
        self.assertEqual(s.get_number_of_nodes(), 3)
        self.assertEqual(s.node_tags, {})
        np.testing.assert_array_equal(s.members, members)
        np.testing.assert_array_equal(s.member_tags['bar'],
                                      np.arange(2, 3, dtype=np.int64))
        np.testing.assert_array_equal(s.member_tags['string'],
                                      np.arange(0, 2, dtype=np.int64))
        self.assertEqual(s.get_number_of_nodes(), 3)
        self.assertEqual(s.get_number_of_members(), 3)
        self.assertEqual(len(s.member_properties), 3)

        member_defaults = Structure.member_defaults.copy()
        Structure.member_defaults['vertical'] = {'linewidth': 1001, 'volume': 2}
        s = Structure(nodes, members,
                      member_tags={
                          'bar': np.arange(1, 3, dtype=np.int64),
                          'string': np.arange(0, 1, dtype=np.int64),
                          'vertical': np.arange(2, 3, dtype=np.int64)
                      },
                      node_tags={
                          'bottom': np.array([0, 1], dtype=np.int64),
                          'top': np.array([2], dtype=np.int64)
                      })
        self.assertEqual(s.label, None)
        self.assertEqual(s.nodes.dtype, np.float_)
        self.assertEqual(s.members.dtype, np.int64)
        np.testing.assert_array_equal(s.nodes, nodes)
        self.assertEqual(len(s.node_properties), 3)
        self.assertEqual(s.get_number_of_nodes(), 3)
        self.assertEqual(set(s.node_tags.keys()), {'bottom', 'top'})
        np.testing.assert_array_equal(s.node_tags['bottom'],
                                      np.array([0, 1], dtype=np.int64))
        np.testing.assert_array_equal(s.node_tags['top'], np.array([2], dtype=np.int64))
        np.testing.assert_array_equal(s.members, members)
        np.testing.assert_array_equal(s.member_tags['bar'],
                                      np.arange(1, 3, dtype=np.int64))
        np.testing.assert_array_equal(s.member_tags['string'],
                                      np.arange(0, 1, dtype=np.int64))
        np.testing.assert_array_equal(s.member_tags['vertical'],
                                      np.arange(2, 3, dtype=np.int64))
        self.assertEqual(s.get_number_of_nodes(), 3)
        self.assertEqual(s.get_number_of_members(), 3)
        self.assertEqual(s.get_member_tags(0), ['string'])
        self.assertEqual(s.get_member_tags(1), ['bar'])
        self.assertEqual(s.get_member_tags(2), ['bar', 'vertical'])
        self.assertTrue(s.has_member_tag(2, 'bar'))
        self.assertTrue(s.has_member_tag(2, 'vertical'))
        self.assertFalse(s.has_member_tag(1, 'vertical'))
        np.testing.assert_array_equal(s.get_members_by_tag('bar'), [1, 2])
        np.testing.assert_array_equal(s.get_members_by_tag('bar', 'vertical'), [2])
        self.assertEqual(len(s.member_properties), 3)
        self.assertEqual(s.member_properties.loc[1, 'linewidth'], 2)
        self.assertEqual(s.member_properties.loc[2, 'linewidth'], 1001)
        self.assertEqual(s.get_member_properties([1, 2],
                                                 'volume', 'mass').to_dict('index'),
                         {1: {'volume': 0., 'mass': 1.}, 2: {'volume': 2., 'mass': 1.}})
        self.assertEqual(s.get_member_properties(2, 'volume', 'mass').to_dict(),
                         {'volume': 2., 'mass': 1.})
        Structure.member_defaults = member_defaults

    def test_get_set_member_properties(self):

        nodes = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        members = np.array([[0, 1, 2], [1, 2, 0]])
        s = Structure(nodes, members,
                      member_tags={
                          'bar': np.arange(1, 3, dtype=np.int64),
                          'string': np.arange(0, 1, dtype=np.int64),
                          'vertical': np.arange(2, 3, dtype=np.int64)
                      },
                      node_tags={
                          'bottom': np.array([0, 1], dtype=np.int64),
                          'top': np.array([2], dtype=np.int64)
                      })

        properties = s.get_member_properties([1, 2], 'volume')
        self.assertIsInstance(properties, pd.Series)
        self.assertEqual(len(properties), 2)
        np.testing.assert_array_equal(properties.values, [0, 0])

        properties = s.get_member_properties(slice(None), 'volume')
        self.assertIsInstance(properties, pd.Series)
        self.assertEqual(len(properties), 3)

        properties = s.get_member_properties([1, 2], 'facecolor')
        self.assertIsInstance(properties, pd.Series)
        self.assertEqual(len(properties), 2)

        properties = s.get_member_properties(slice(None), 'facecolor')
        self.assertIsInstance(properties, pd.Series)
        self.assertEqual(len(properties), 3)

        properties = s.get_member_properties([1, 2], 'volume', 'facecolor')
        self.assertIsInstance(properties, pd.DataFrame)
        self.assertEqual(len(properties), 2)

        properties = s.get_member_properties(slice(None), 'volume', 'facecolor')
        self.assertIsInstance(properties, pd.DataFrame)
        self.assertEqual(len(properties), 3)

        s.set_member_properties(2, 'volume', 3)
        np.testing.assert_array_equal(s.member_properties['volume'].values, [0, 0, 3])

        s.set_member_properties([1, 2], 'volume', [1, 2], scalar=False)
        np.testing.assert_array_equal(s.member_properties['volume'].values, [0, 1, 2])

        s.set_member_properties([1, 2], 'volume', 3)
        np.testing.assert_array_equal(s.member_properties['volume'].values, [0, 3, 3])

        s.set_member_properties([1, 2], 'volume', 1)
        np.testing.assert_array_equal(s.member_properties['volume'].values, [0, 1, 1])

        s.set_member_properties([1, 2], 'volume', 2, 'mass', 4)
        np.testing.assert_array_equal(s.member_properties['volume'].values, [0, 2, 2])
        np.testing.assert_array_equal(s.member_properties['mass'].values, [1, 4, 4])

        s.set_member_properties(2, 'facecolor', utils.Colors.BROWN.value)
        self.assertEqual(s.member_properties['facecolor'].tolist(),
                         [utils.Colors.ORANGE.value,
                          utils.Colors.BLUE.value, utils.Colors.BROWN.value])

        s.set_member_properties([1, 2], 'facecolor', utils.Colors.GREEN.value)
        self.assertEqual(s.member_properties['facecolor'].tolist(),
                         [utils.Colors.ORANGE.value,
                          utils.Colors.GREEN.value, utils.Colors.GREEN.value])

        s.set_member_properties([1, 2], 'facecolor',
                                [utils.Colors.BLUE.value, utils.Colors.ORANGE.value],
                                scalar=False)
        self.assertEqual(s.member_properties['facecolor'].tolist(),
                         [utils.Colors.ORANGE.value,
                          utils.Colors.BLUE.value, utils.Colors.ORANGE.value])

    def test_add_members_and_add_nodes(self):

        nodes1 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        members1 = np.array([[0, 1, 2], [1, 2, 0]])
        s = Structure(nodes1, members1,
                      member_tags={
                          'bar': np.arange(1, 3, dtype=np.int64),
                          'string': np.arange(0, 1, dtype=np.int64),
                          'vertical': np.arange(2, 3, dtype=np.int64)
                      })
        members2 = s.get_number_of_nodes() + np.array([[0], [1]])
        nodes2 = np.array([[1, 0], [1, 1], [0, 1]])
        s.add_nodes(nodes2, node_tags={
            'bottom': np.array([0], dtype=np.int64),
            'top': np.array([1], dtype=np.int64)
        })
        s.add_members(members2, number_of_strings=1)

        self.assertEqual(s.get_number_of_nodes(), 5)
        self.assertEqual(len(s.node_properties), 5)
        self.assertEqual(set(s.node_tags.keys()), {'bottom', 'top'})
        np.testing.assert_array_equal(s.node_tags['bottom'],
                                      np.array([3], dtype=np.int64))
        np.testing.assert_array_equal(s.node_tags['top'],
                                      np.array([4], dtype=np.int64))
        self.assertEqual(s.get_number_of_members(), 4)
        self.assertEqual(set(s.member_tags.keys()), {'bar', 'string', 'vertical'})
        np.testing.assert_array_equal(s.member_tags['bar'], [1, 2])
        np.testing.assert_array_equal(s.member_tags['string'], [0, 3])
        np.testing.assert_array_equal(s.member_tags['vertical'], [2])
        np.testing.assert_array_equal(s.members, [[0, 1, 2, 3], [1, 2, 0, 4]])
        np.testing.assert_array_equal(s.nodes,
                                      [[1, 0, 0, 1, 0],
                                       [0, 1, 0, 1, 1],
                                       [0, 0, 1, 0, 1]])
        np.testing.assert_array_equal(s.member_properties.index, np.arange(4))

    def test_copy(self):

        nodes = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        members = np.array([[0, 1, 2], [1, 2, 0]])
        member_defaults = Structure.member_defaults.copy()
        Structure.member_defaults['vertical'] = {'linewidth': 1001, 'volume': 2}
        s = Structure(nodes, members,
                      member_tags={
                          'bar': np.arange(1, 3, dtype=np.int64),
                          'string': np.arange(0, 1, dtype=np.int64),
                          'vertical': np.arange(2, 3, dtype=np.int64)
                      },
                      node_tags={
                          'bottom': np.array([0, 1], dtype=np.int64),
                          'top': np.array([2], dtype=np.int64)
                      }, label='label')

        self.assertEqual(s.label, 'label')
        self.assertEqual(s.nodes.dtype, np.float_)
        self.assertEqual(s.members.dtype, np.int64)
        np.testing.assert_array_equal(s.nodes, nodes)
        self.assertEqual(len(s.node_properties), 3)
        self.assertEqual(s.get_number_of_nodes(), 3)
        self.assertEqual(set(s.node_tags.keys()), {'bottom', 'top'})
        np.testing.assert_array_equal(s.node_tags['bottom'],
                                      np.array([0, 1], dtype=np.int64))
        np.testing.assert_array_equal(s.node_tags['top'],
                                      np.array([2], dtype=np.int64))
        np.testing.assert_array_equal(s.members, members)
        np.testing.assert_array_equal(s.member_tags['bar'],
                                      np.arange(1, 3, dtype=np.int64))
        np.testing.assert_array_equal(s.member_tags['string'],
                                      np.arange(0, 1, dtype=np.int64))
        np.testing.assert_array_equal(s.member_tags['vertical'],
                                      np.arange(2, 3, dtype=np.int64))
        self.assertEqual(s.get_number_of_nodes(), 3)
        self.assertEqual(s.get_number_of_members(), 3)
        self.assertEqual(s.get_member_tags(0), ['string'])
        self.assertEqual(s.get_member_tags(1), ['bar'])
        self.assertEqual(s.get_member_tags(2), ['bar', 'vertical'])
        self.assertTrue(s.has_member_tag(2, 'bar'))
        self.assertTrue(s.has_member_tag(2, 'vertical'))
        self.assertFalse(s.has_member_tag(1, 'vertical'))
        np.testing.assert_array_equal(s.get_members_by_tag('bar'), [1, 2])
        np.testing.assert_array_equal(s.get_members_by_tag('bar', 'vertical'), [2])
        self.assertEqual(len(s.member_properties), 3)
        self.assertEqual(s.member_properties.loc[1, 'linewidth'], 2)
        self.assertEqual(s.member_properties.loc[2, 'linewidth'], 1001)
        self.assertEqual(s.get_member_properties([1, 2],
                                                 'volume', 'mass').to_dict('index'),
                         {1: {'volume': 0., 'mass': 1.}, 2: {'volume': 2., 'mass': 1.}})
        self.assertEqual(s.get_member_properties(2, 'volume', 'mass').to_dict(),
                         {'volume': 2., 'mass': 1.})

        copy = s.copy()

        self.assertEqual(copy.label, 'label')
        self.assertEqual(copy.nodes.dtype, np.float_)
        self.assertEqual(copy.members.dtype, np.int64)
        np.testing.assert_array_equal(copy.nodes, nodes)
        self.assertEqual(len(copy.node_properties), 3)
        self.assertEqual(copy.get_number_of_nodes(), 3)
        self.assertEqual(set(copy.node_tags.keys()), {'bottom', 'top'})
        np.testing.assert_array_equal(copy.node_tags['bottom'],
                                      np.array([0, 1], dtype=np.int64))
        np.testing.assert_array_equal(copy.node_tags['top'],
                                      np.array([2], dtype=np.int64))
        np.testing.assert_array_equal(copy.members, members)
        np.testing.assert_array_equal(copy.member_tags['bar'],
                                      np.arange(1, 3, dtype=np.int64))
        np.testing.assert_array_equal(copy.member_tags['string'],
                                      np.arange(0, 1, dtype=np.int64))
        np.testing.assert_array_equal(copy.member_tags['vertical'],
                                      np.arange(2, 3, dtype=np.int64))
        self.assertEqual(copy.get_number_of_nodes(), 3)
        self.assertEqual(copy.get_number_of_members(), 3)
        self.assertEqual(copy.get_member_tags(0), ['string'])
        self.assertEqual(copy.get_member_tags(1), ['bar'])
        self.assertEqual(copy.get_member_tags(2), ['bar', 'vertical'])
        self.assertTrue(copy.has_member_tag(2, 'bar'))
        self.assertTrue(copy.has_member_tag(2, 'vertical'))
        self.assertFalse(copy.has_member_tag(1, 'vertical'))
        np.testing.assert_array_equal(copy.get_members_by_tag('bar'), [1, 2])
        np.testing.assert_array_equal(copy.get_members_by_tag('bar', 'vertical'), [2])
        self.assertEqual(len(copy.member_properties), 3)
        self.assertEqual(copy.member_properties.loc[1, 'linewidth'], 2)
        self.assertEqual(copy.member_properties.loc[2, 'linewidth'], 1001)
        self.assertEqual(copy.get_member_properties([1, 2],
                                                    'volume', 'mass').to_dict('index'),
                         {1: {'volume': 0., 'mass': 1.}, 2: {'volume': 2., 'mass': 1.}})
        self.assertEqual(copy.get_member_properties(2, 'volume', 'mass').to_dict(),
                         {'volume': 2., 'mass': 1.})

        Structure.member_defaults = member_defaults

    def test_merge(self):

        nodes1 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        members1 = np.array([[0, 1, 2], [1, 2, 0]])
        s1 = Structure(nodes1, members1,
                       member_tags={
                           'bar': np.arange(1, 3, dtype=np.int64),
                           'string': np.arange(0, 1, dtype=np.int64),
                           'vertical': np.arange(2, 3, dtype=np.int64)
                       },
                       node_tags={
                           'bottom': np.array([0, 1], dtype=np.int64),
                           'top': np.array([2], dtype=np.int64)
                       })

        nodes2 = np.array([[1, 0], [1, 1], [0, 1]])
        members2 = np.array([[0], [1]])
        s2 = Structure(nodes2, members2, number_of_strings=1)

        s1.merge(s2)

        self.assertEqual(s1.get_number_of_nodes(), 5)
        self.assertEqual(len(s1.node_properties), 5)
        self.assertEqual(set(s1.node_tags.keys()), {'bottom', 'top'})
        np.testing.assert_array_equal(s1.node_tags['bottom'],
                                      np.array([0, 1], dtype=np.int64))
        np.testing.assert_array_equal(s1.node_tags['top'],
                                      np.array([2], dtype=np.int64))
        self.assertEqual(s1.get_number_of_members(), 4)
        self.assertEqual(set(s1.member_tags.keys()), {'bar', 'string', 'vertical'})
        np.testing.assert_array_equal(s1.member_tags['bar'], [1, 2])
        np.testing.assert_array_equal(s1.member_tags['string'], [0, 3])
        np.testing.assert_array_equal(s1.member_tags['vertical'], [2])
        np.testing.assert_array_equal(s1.members, [[0, 1, 2, 3], [1, 2, 0, 4]])
        np.testing.assert_array_equal(s1.nodes,
                                      [[1, 0, 0, 1, 0],
                                       [0, 1, 0, 1, 1],
                                       [0, 0, 1, 0, 1]])
        np.testing.assert_array_equal(s1.member_properties.index, np.arange(4))

        # nodes1 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        # members1 = np.array([[0, 1, 2], [1, 2, 0]])
        s1 = Structure(nodes1, members1,
                       member_tags={
                           'bar': np.arange(1, 3, dtype=np.int64),
                           'string': np.arange(0, 1, dtype=np.int64),
                           'vertical': np.arange(2, 3, dtype=np.int64)
                       },
                       node_tags={
                           'bottom': np.array([0, 1], dtype=np.int64),
                           'top': np.array([2], dtype=np.int64)
                       })

        # nodes2 = np.array([[1, 0], [1, 1], [0, 1]])
        # members2 = np.array([[0], [1]])
        s2 = Structure(nodes2, members2, number_of_strings=1)

        s2.merge(s1)
        self.assertEqual(s2.get_number_of_nodes(), 5)
        self.assertEqual(len(s2.node_properties), 5)
        self.assertEqual(set(s2.node_tags.keys()), {'bottom', 'top'})
        np.testing.assert_array_equal(s2.node_tags['bottom'],
                                      np.array([2, 3], dtype=np.int64))
        np.testing.assert_array_equal(s2.node_tags['top'],
                                      np.array([4], dtype=np.int64))
        self.assertEqual(s2.get_number_of_members(), 4)
        self.assertEqual(set(s2.member_tags.keys()), {'bar', 'string', 'vertical'})
        np.testing.assert_array_equal(s2.member_tags['bar'], [2, 3])
        np.testing.assert_array_equal(s2.member_tags['string'], [0, 1])
        np.testing.assert_array_equal(s2.member_tags['vertical'], [3])
        np.testing.assert_array_equal(s2.members, [[0, 2, 3, 4], [1, 3, 4, 2]])
        np.testing.assert_array_equal(s2.nodes,
                                      [[1, 0, 1, 0, 0],
                                       [1, 1, 0, 1, 0],
                                       [0, 1, 0, 0, 1]])
        np.testing.assert_array_equal(s2.member_properties.index, np.arange(4))

        s1 = Structure(nodes1, members1,
                       member_tags={
                           'bar': np.arange(1, 3, dtype=np.int64),
                           'string': np.arange(0, 1, dtype=np.int64),
                           'vertical': np.arange(2, 3, dtype=np.int64)
                       },
                       node_tags={
                           'bottom': np.array([0, 1], dtype=np.int64),
                           'top': np.array([2], dtype=np.int64)
                       })

        # nodes2 = np.array([[1, 0], [1, 1], [0, 1]])
        # members2 = np.array([[0], [1]])
        s2 = Structure(nodes2, members2, number_of_strings=1)

        s = structure.merge(s1, s2)

        self.assertFalse(s is s1)
        self.assertFalse(s is s2)
        self.assertEqual(s.get_number_of_nodes(), 5)
        self.assertEqual(len(s.node_properties), 5)
        self.assertEqual(set(s.node_tags.keys()), {'bottom', 'top'})
        np.testing.assert_array_equal(s.node_tags['bottom'],
                                      np.array([0, 1], dtype=np.int64))
        np.testing.assert_array_equal(s.node_tags['top'],
                                      np.array([2], dtype=np.int64))
        self.assertEqual(s.get_number_of_members(), 4)
        self.assertEqual(set(s.member_tags.keys()), {'bar', 'string', 'vertical'})
        np.testing.assert_array_equal(s.member_tags['bar'], [1, 2])
        np.testing.assert_array_equal(s.member_tags['string'], [0, 3])
        np.testing.assert_array_equal(s.member_tags['vertical'], [2])
        np.testing.assert_array_equal(s.members, [[0, 1, 2, 3], [1, 2, 0, 4]])
        np.testing.assert_array_equal(s.nodes,
                                      [[1, 0, 0, 1, 0],
                                       [0, 1, 0, 1, 1],
                                       [0, 0, 1, 0, 1]])
        np.testing.assert_array_equal(s.member_properties.index, np.arange(4))

    def test_length_com_cog_translate(self):

        nodes1 = np.array([[1, 0, 0, 0], [0, 1, 0, 1], [0, 0, 2, 1]])
        members1 = np.array([[0, 1, 2, 3], [1, 2, 0, 0]])
        s = Structure(nodes1, members1, number_of_strings=2)

        # test get member length
        np.testing.assert_array_equal(s.get_member_length(),
                                      [np.linalg.norm(nodes1[:, 1]-nodes1[:, 0]),
                                       np.linalg.norm(nodes1[:, 2]-nodes1[:, 1]),
                                       np.linalg.norm(nodes1[:, 0]-nodes1[:, 2]),
                                       np.linalg.norm(nodes1[:, 3]-nodes1[:, 0])])

        # test center of mass
        np.testing.assert_array_equal(s.get_center_of_mass(),
                                      (nodes1[:, 1] + nodes1[:, 0] +
                                       nodes1[:, 2] + nodes1[:, 1] +
                                       nodes1[:, 0] + nodes1[:, 2] +
                                       nodes1[:, 3] + nodes1[:, 0])/8)

        # test center of gravity
        np.testing.assert_array_equal(s.get_centroid(),
                                      (nodes1[:, 0] + nodes1[:, 1] +
                                       nodes1[:, 2] + nodes1[:, 3])/4)

    def test_rotate(self):

        nodes1 = np.array([[1, 0, 0, 0], [0, 1, 0, 1], [0, 0, 2, 1]])
        members1 = np.array([[0, 1, 2, 3], [1, 2, 0, 0]])
        s = Structure(nodes1, members1, number_of_strings=2)

        v = np.array([1, 1, 1])
        s.rotate(v)
        R = scipy.spatial.transform.Rotation.from_rotvec(v)
        np.testing.assert_array_equal(s.nodes, R.apply(nodes1.transpose()).transpose())

    def test_reflect(self):

        nodes1 = np.array([[1, 0, 0, 0], [0, 1, 0, 1], [0, 0, 2, 1]])
        members1 = np.array([[0, 1, 2, 3], [1, 2, 0, 0]])
        s = Structure(nodes1, members1, number_of_strings=2)

        v = np.array([1, 1, 1])
        s.reflect(v)
        H = np.eye(3) - 2 * np.outer(v, v)/(np.linalg.norm(v)**2)
        np.testing.assert_array_equal(s.nodes, H @ nodes1)

        s.set_nodes(nodes1)
        p = np.array([1, -1, 2])
        s.reflect(v, p)
        p = p.reshape((3, 1))
        np.testing.assert_array_equal(s.nodes, (H @ (nodes1 - p)) + p)

    def test_remove_nodes(self):

        nodes1 = np.array([[1, 0, 0, 0], [0, 1, 0, 1], [0, 0, 2, 1]])
        nodes1_tags = {'tags': np.array([0, 2, 3], dtype=np.int64)}
        members1 = np.array([[0, 1, 2], [1, 2, 0]])
        s = Structure(nodes1, members1, number_of_strings=2, node_tags=nodes1_tags)

        np.testing.assert_array_equal(s.get_unused_nodes(), [3])
        self.assertTrue(s.has_unused_nodes())

        s.remove_nodes([3])
        np.testing.assert_array_equal(s.nodes, nodes1[:, :3])
        self.assertEqual(len(s.node_properties), 3)
        np.testing.assert_array_equal(s.node_properties.index, np.arange(3))
        np.testing.assert_array_equal(s.members, members1)
        np.testing.assert_array_equal(s.node_tags['tags'], [0, 2])
        self.assertFalse(s.has_unused_nodes())

        members2 = np.array([[0, 1, 3], [1, 3, 0]])
        s = Structure(nodes1, members2, number_of_strings=2, node_tags=nodes1_tags)

        np.testing.assert_array_equal(s.get_unused_nodes(), [2])
        self.assertTrue(s.has_unused_nodes())

        s.remove_nodes([2])
        np.testing.assert_array_equal(s.nodes, nodes1[:, [0, 1, 3]])
        self.assertEqual(len(s.node_properties), 3)
        np.testing.assert_array_equal(s.node_properties.index, np.arange(3))
        np.testing.assert_array_equal(s.node_tags['tags'], [0, 2])
        np.testing.assert_array_equal(s.members, members1)

        members3 = np.array([[0, 3], [3, 0]])
        s = Structure(nodes1, members3, number_of_strings=2, node_tags=nodes1_tags)

        np.testing.assert_array_equal(s.get_unused_nodes(), [1, 2])
        self.assertTrue(s.has_unused_nodes())

        s.remove_nodes([1, 2])
        np.testing.assert_array_equal(s.nodes, nodes1[:, [0, 3]])
        np.testing.assert_array_equal(s.members, np.array([[0, 1], [1, 0]]))
        np.testing.assert_array_equal(s.node_tags['tags'], [0, 1])
        self.assertFalse(s.has_unused_nodes())

        members4 = np.array([[1, 3], [3, 1]])
        s = Structure(nodes1, members4, number_of_strings=2, node_tags=nodes1_tags)

        np.testing.assert_array_equal(s.get_unused_nodes(), [0, 2])
        self.assertTrue(s.has_unused_nodes())

        s.remove_nodes([0, 2])
        np.testing.assert_array_equal(s.nodes, nodes1[:, [1, 3]])
        self.assertEqual(len(s.node_properties), 2)
        np.testing.assert_array_equal(s.node_properties.index, np.arange(2))
        np.testing.assert_array_equal(s.members, np.array([[0, 1], [1, 0]]))
        np.testing.assert_array_equal(s.node_tags['tags'], [1])
        self.assertFalse(s.has_unused_nodes())

        # has used nodes
        s = Structure(nodes1, members1, number_of_strings=2, node_tags=nodes1_tags)

        np.testing.assert_array_equal(s.get_unused_nodes(), [3])
        self.assertTrue(s.has_unused_nodes())

        with self.assertWarns(Warning):
            s.remove_nodes([2, 3])

        np.testing.assert_array_equal(s.nodes, nodes1[:, :3])
        self.assertEqual(len(s.node_properties), 3)
        np.testing.assert_array_equal(s.node_properties.index, np.arange(3))
        np.testing.assert_array_equal(s.members, members1)
        np.testing.assert_array_equal(s.node_tags['tags'], [0, 2])
        self.assertFalse(s.has_unused_nodes())

        s = Structure(nodes1, members1, number_of_strings=2, node_tags=nodes1_tags)

        np.testing.assert_array_equal(s.get_unused_nodes(), [3])
        self.assertTrue(s.has_unused_nodes())

        with self.assertWarns(Warning):
            s.remove_nodes([1, 2])

        np.testing.assert_array_equal(s.nodes, nodes1)
        self.assertEqual(len(s.node_properties), 4)
        np.testing.assert_array_equal(s.node_properties.index, np.arange(4))
        np.testing.assert_array_equal(s.members, members1)
        np.testing.assert_array_equal(s.node_tags['tags'], [0, 2, 3])
        self.assertTrue(s.has_unused_nodes())

        # None as parameter

        members3 = np.array([[0, 3], [3, 0]])
        s = Structure(nodes1, members3, number_of_strings=2, node_tags=nodes1_tags)

        np.testing.assert_array_equal(s.get_unused_nodes(), [1, 2])
        self.assertTrue(s.has_unused_nodes())

        s.remove_nodes()
        np.testing.assert_array_equal(s.nodes, nodes1[:, [0, 3]])
        self.assertEqual(len(s.node_properties), 2)
        np.testing.assert_array_equal(s.node_properties.index, np.arange(2))
        np.testing.assert_array_equal(s.members, np.array([[0, 1], [1, 0]]))
        np.testing.assert_array_equal(s.node_tags['tags'], [0, 1])
        self.assertFalse(s.has_unused_nodes())

    def test_remove_members(self):

        nodes = np.array([[1, 0, 0, 0], [0, 1, 0, 1], [0, 0, 2, 1]])
        nodes_tags = {'tags': np.array([0, 2, 3], dtype=np.int64)}
        members = np.array([[0, 1, 2], [1, 2, 0]])
        members_tag = {'mtag': np.array([0, 2], dtype=np.int64)}

        s = Structure(nodes, members, number_of_strings=2,
                      node_tags=nodes_tags, member_tags=members_tag)

        s.remove_members([2])
        np.testing.assert_array_equal(s.nodes, nodes)
        self.assertEqual(len(s.node_properties), 4)
        np.testing.assert_array_equal(s.members, members[:, :2])
        np.testing.assert_array_equal(s.member_tags['mtag'], [0])

        s = Structure(nodes, members, number_of_strings=2,
                      node_tags=nodes_tags, member_tags=members_tag)

        s.remove_members([0, 1])
        np.testing.assert_array_equal(s.nodes, nodes)
        self.assertEqual(len(s.node_properties), 4)
        np.testing.assert_array_equal(s.members, members[:, [2]])
        np.testing.assert_array_equal(s.member_tags['mtag'], [0])

        s = Structure(nodes, members, number_of_strings=2,
                      node_tags=nodes_tags, member_tags=members_tag)

        s.remove_members([1, 2])
        np.testing.assert_array_equal(s.nodes, nodes)
        self.assertEqual(len(s.node_properties), 4)
        np.testing.assert_array_equal(s.members, members[:, [0]])
        np.testing.assert_array_equal(s.member_tags['mtag'], [0])

    def test_close_node(self):

        nodes1 = np.array([[1, 0, 0, 0], [0, 1, 0, 1], [0, 0, 2, 1]])
        members1 = np.array([[0, 1, 2], [1, 2, 0]])
        s = Structure(nodes1, members1, number_of_strings=2)
        merge, merge_map = s.get_close_nodes()
        self.assertFalse(merge)

        s.merge_close_nodes()
        self.assertEqual(s.get_number_of_nodes(), 4)
        np.testing.assert_array_equal(s.nodes, nodes1)
        np.testing.assert_array_equal(s.members, members1)

        nodes1 = np.array([[1, 0, 0, 1], [0, 1, 0, 0], [0, 0, 2, 0]])
        members1 = np.array([[0, 1, 2, 3], [1, 2, 0, 1]])
        s = Structure(nodes1, members1, number_of_strings=2)
        merge, merge_map = s.get_close_nodes()
        self.assertEqual(merge, {3})
        np.testing.assert_array_equal(merge_map, [0, 1, 2, 0])

        s.merge_close_nodes()
        self.assertEqual(s.get_number_of_nodes(), 3)
        np.testing.assert_array_equal(s.nodes, nodes1[:, :3])
        np.testing.assert_array_equal(s.members, np.array([[0, 1, 2, 0], [1, 2, 0, 1]]))

        nodes1 = np.array([[1, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 2]])
        members1 = np.array([[0, 1, 2, 3], [1, 2, 0, 1]])
        s = Structure(nodes1, members1, number_of_strings=2)
        merge, merge_map = s.get_close_nodes()
        self.assertEqual(merge, {2})
        np.testing.assert_array_equal(merge_map, [0, 1, 0, 3])

        s.merge_close_nodes()
        self.assertEqual(s.get_number_of_nodes(), 3)
        np.testing.assert_array_equal(s.nodes, nodes1[:, [0, 1, 3]])
        np.testing.assert_array_equal(s.members, np.array([[0, 1, 0, 2], [1, 0, 0, 1]]))

        radius = 0.1
        nodes1 = np.array([[1, 0, 0, 1, 1],
                           [0, 1, 0, 0, 0],
                           [0, 0, 2, .9*radius, 1.8*radius]])
        members1 = np.array([[0, 1, 2, 3, 2], [1, 2, 0, 1, 4]])
        s = Structure(nodes1, members1, number_of_strings=2)
        merge, merge_map = s.get_close_nodes(radius)
        self.assertEqual(merge, {3, 4})
        np.testing.assert_array_equal(merge_map, [0, 1, 2, 0, 0])

        s.merge_close_nodes(radius)
        self.assertEqual(s.get_number_of_nodes(), 3)
        np.testing.assert_array_equal(s.nodes, nodes1[:, :3])
        np.testing.assert_array_equal(s.members,
                                      np.array([[0, 1, 2, 0, 2], [1, 2, 0, 1, 0]]))

    def test_equilibrium_update_member_properties(self):

        nodes = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0]])
        members = np.array([[0, 1], [1, 2], [2, 0], [0, 3], [1, 3], [2, 3]]).transpose()
        s0 = Structure(nodes, members)

        center = s0.get_centroid()
        s0.add_nodes(center.reshape((3, 1)))

        s1 = s0.copy()

        members = np.array([[0, 4], [1, 4], [2, 4]]).transpose()
        s1.add_members(members, number_of_strings=3)

        # won't be in equilibrium with only three strings
        self.assertRaises(Exception, s1.equilibrium)

        s1 = s0.copy()

        members = np.array([[0, 4], [1, 4], [2, 4], [3, 4]]).transpose()
        s1.add_members(members, number_of_strings=4)

        # equilibrium
        s1.equilibrium()

        # set properties
        s1.update_member_properties()

        # forces
        force = np.hstack((np.kron(np.array([-np.sqrt(2), -1, np.sqrt(11)]),
                                   np.ones((3,))), [np.sqrt(3)]))
        np.testing.assert_allclose(s1.member_properties['force'], force)

        lambda_ = np.hstack((np.kron(np.array([-1, -1, 4]), np.ones((3,))), [4]))
        np.testing.assert_allclose(s1.member_properties['lambda_'], lambda_)

        mass = np.hstack((157./200. * np.ones((6,)), 157./200./4 * np.ones((4,))))
        np.testing.assert_allclose(
            s1.member_properties['mass'] / s1.get_member_length() / np.pi,
            mass)

        volume = np.hstack((np.ones((6,)), 1/4 * np.ones((4,))))
        np.testing.assert_allclose(
            10000 * s1.member_properties['volume'] / s1.get_member_length() / np.pi,
            volume)

        np.testing.assert_allclose(
            s1.member_properties['stiffness'],
            np.pi * s1.member_properties['modulus'] *
            (s1.member_properties['radius']**2 -
             s1.member_properties['inner_radius']**2)
            / s1.get_member_length())

        np.testing.assert_allclose(
            s1.member_properties['rest_length'],
            s1.get_member_length() * (1 - s1.member_properties['lambda_'].values
                                      / s1.member_properties['stiffness']))

    def test_node_tags(self):

        nodes = np.array([[1, 0, 0, 0, 1], [0, 1, 0, 1, 1], [0, 0, 2, 1, 3]])
        nodes_tags = {
            'nags0': np.array([0, 3], dtype=np.int64),
            'nags1': np.array([1, 2], dtype=np.int64)
        }
        members = np.array([[0, 1, 2, 3], [1, 2, 0, 4]])
        members_tag = {
            'mags0': np.array([0], dtype=np.int64),
            'mags1': np.array([1, 2], dtype=np.int64)
        }
        s = Structure(nodes, members,
                      number_of_strings=2,
                      node_tags=nodes_tags, member_tags=members_tag)

        s.add_node_tag('nags2', [0, 3])
        np.testing.assert_array_equal(s.node_tags['nags2'], [0, 3])

        s.add_node_tag('nags1', [3])
        np.testing.assert_array_equal(s.node_tags['nags1'], [1, 2, 3])

        s.remove_node_tag('nags1', [2])
        np.testing.assert_array_equal(s.node_tags['nags1'], [1, 3])

        s.delete_node_tag('nags1')
        self.assertFalse('nags1' in s.node_tags)

    def test_member_tags(self):

        nodes = np.array([[1, 0, 0, 0, 1], [0, 1, 0, 1, 1], [0, 0, 2, 1, 3]])
        nodes_tags = {
            'nags0': np.array([0, 3], dtype=np.int64),
            'nags1': np.array([1, 2], dtype=np.int64)
        }
        members = np.array([[0, 1, 2, 3], [1, 2, 0, 4]])
        members_tag = {
            'mags0': np.array([0], dtype=np.int64),
            'mags1': np.array([1, 2], dtype=np.int64)
        }
        s = Structure(nodes, members,
                      number_of_strings=2,
                      node_tags=nodes_tags, member_tags=members_tag)

        s.add_member_tag('mags2', [0, 3])
        np.testing.assert_array_equal(s.member_tags['mags2'], [0, 3])

        s.add_member_tag('mags1', [3])
        np.testing.assert_array_equal(s.member_tags['mags1'], [1, 2, 3])

        s.remove_member_tag('mags1', [2])
        np.testing.assert_array_equal(s.member_tags['mags1'], [1, 3])

        s.delete_member_tag('mags1')
        self.assertFalse('mags1' in s.member_tags)

    def test_node_constraint(self):

        nodes = np.array([[1, 0, 0, 0, 1], [0, 1, 0, 1, 1], [0, 0, 2, 1, 3]])
        nodes_tags = {
            'nags0': np.array([0, 3], dtype=np.int64),
            'nags1': np.array([1, 2], dtype=np.int64)
        }
        members = np.array([[0, 1, 2, 3], [1, 2, 0, 4]])
        members_tag = {
            'mags0': np.array([0], dtype=np.int64),
            'mags1': np.array([1, 2], dtype=np.int64)
        }
        s = Structure(nodes, members,
                      number_of_strings=2,
                      node_tags=nodes_tags, member_tags=members_tag)
        s.node_properties.loc[2, 'constraint'] = NodeConstraint()
        s.node_properties.loc[0, 'constraint'] = NodeConstraint(np.array(([[1, 1, 0]])))

        # print(s.node_properties.loc[2, 'constraint'])
        # print(s.node_properties.loc[2, 'constraint'].__class__)
        # print(s.node_properties.loc[0, 'constraint'])
        # print(s.node_properties.loc[0, 'constraint'].__class__)

        R, T = NodeConstraint.node_constraint(s.nodes, s.node_properties['constraint'])
        self.assertEqual(R.shape, (4, 15))
        self.assertEqual(T.shape, (15, 11))

    def test_dependent_nodes(self):

        nodes = np.array([[1, 0, 0, 0, 1, 1/2],
                          [0, 1, 0, 1, 1, 1/2],
                          [0, 0, 2, 1, 3, 0]])
        nodes_tags = {
            'nags0': np.array([0, 3], dtype=np.int64),
            'nags1': np.array([1, 2], dtype=np.int64)
        }
        members = np.array([[0, 1, 2, 3, 0, 1], [1, 2, 0, 4, 5, 5]])
        members_tag = {
            'mags0': np.array([0], dtype=np.int64),
            'mags1': np.array([1, 2, 5], dtype=np.int64)
        }
        s = Structure(nodes, members,
                      number_of_strings=2,
                      node_tags=nodes_tags, member_tags=members_tag)
        np.testing.assert_array_equal(s.get_members_per_node(), [3, 3, 2, 1, 1, 2])
        np.testing.assert_array_equal(s.get_colinear_nodes(), [5])
        dep_nodes, dep_members = s.get_colinear_nodes(return_members=True)
        np.testing.assert_array_equal(dep_nodes, [5])
        np.testing.assert_array_equal(dep_members, [[4, 5]])

        s.merge_colinear_nodes()
        self.assertEqual(s.get_number_of_nodes(), 5)
        self.assertEqual(s.get_number_of_members(), 5)
        np.testing.assert_array_equal(s.nodes, nodes[:, :-1])
        np.testing.assert_array_equal(s.members,
                                      np.array([[0, 1, 2, 3, 0], [1, 2, 0, 4, 1]]))
        np.testing.assert_array_equal(s.member_tags['bar'], [2, 3, 4])
        np.testing.assert_array_equal(s.member_tags['mags1'], [1, 2, 4])

    def test_slack_members(self):

        nodes = np.array([[1, 0, 0, 0, 1, 1/2],
                          [0, 1, 0, 1, 1, 1/2], [0, 0, 2, 1, 3, 0]])
        nodes_tags = {
            'nags0': np.array([0, 3], dtype=np.int64),
            'nags1': np.array([1, 2], dtype=np.int64)
        }
        members = np.array([[0, 1, 2, 3, 0, 1], [1, 2, 0, 4, 5, 5]])
        members_tag = {
            'mags0': np.array([0], dtype=np.int64),
            'mags1': np.array([1, 2, 5], dtype=np.int64)
        }
        s = Structure(nodes, members,
                      number_of_strings=2,
                      node_tags=nodes_tags, member_tags=members_tag)
        s.member_properties.loc[[0, 3, 5], 'lambda_'] = [-1, 2, -3]
        np.testing.assert_array_equal(s.get_slack_members(), [1, 2, 4])
        s.set_member_properties([2], 'lambda_', 1e-9)
        np.testing.assert_array_equal(s.get_slack_members(), [1, 2, 4])
        s.set_member_properties([0], 'lambda_', 1e-9)
        s.remove_members(s.get_slack_members())
        self.assertEqual(s.get_number_of_members(), 2)


if __name__ == '__main__':
    unittest.main()



from enum import Enum
from typing import Tuple, Union
from typing_extensions import Literal

import numpy as np
import numpy.typing as npt
import scipy


def rotation_2d(phi: float) -> npt.NDArray:
    """
    Return a 2D rotation matrix

    :param phi: the rotation angle in radians
    :return: the rotation matrix
    """
    return np.array([[np.cos(phi), -np.sin(phi)], [np.sin(phi), np.cos(phi)]])


def rotation_3d(v: npt.NDArray) -> npt.NDArray:
    """
    Return a 3D rotation matrix

    :param v: the vector around which to rotate; its norm is the angle to rotate
    :return: the rotation matrix

    **Notes:**

    1. See :meth:`scipy.spatial.transform.Rotation.from_rotvec` for details
    """
    return scipy.spatial.transform.Rotation.from_rotvec(v).as_matrix()


def orthogonalize(a: npt.NDArray, epsilon: float = 1e-8,
                  mode: Literal['reduced', 'complete'] = 'reduced')\
        -> Union[Tuple[int, npt.NDArray, npt.NDArray],Tuple[int, npt.NDArray]]:
    """
    Ortoghonalize the constraint

    .. math::
        A^T \\, x = 0

    :param a: the coefficient array :math:`A`
    :param epsilon: the accuracy with which to evaluate the rank
    :param mode: 'complete' or 'reduced'
    :return: tuple with rank, the orthogonalized coefficient array, and its null space
             if mode is 'complete'

    **Notes:**

        1. If ``mode = 'reduced'`` the constraint coefficient is normalized at the
           constructor by calculating the *reduced* QR decomposition

           .. math::
               A = Q R, \\quad Q^T Q = I, \\quad R \\text{ is upper triangular}

           Assuming that :math:`A` is full rank, the equivalent orthogonal constraint

           .. math::
               A^T x = R^T V^T x = 0 \\quad \\Leftrightarrow \\quad V^T x = 0

           is obtained in which the coefficient is the orthogonal matrix :math:`V = Q`

        2. If ``mode = 'complete'`` the constraint coefficient is normalized at the
           constructor by calculating the *complete* QR decomposition

           .. math::
               A = Q R, \\quad Q^T Q = Q Q^T = I, \\quad R \\text{ is upper triangular}

           Assuming that :math:`A` is full rank, partition

           .. math::
               \\begin{bmatrix} V & T \\end{bmatrix} = Q =
               \\begin{bmatrix} Q_1 & Q_2 \\end{bmatrix}, \\qquad R =
               \\begin{bmatrix} R_1 \\\\ 0 \\end{bmatrix}, \\quad
               R_1 \\text{ is upper triangular}

           to obtain the equivalent orthogonal constraint and its solution

           .. math::
               A^T x = R_1^T V^T x = 0 \\quad \\Leftrightarrow \\quad V^T x = 0,
               \\qquad x = T y

           The matrix :math:`T` is an orthogonal basis for the constraint null space

        3. The above factorizations are modified to take into account the numerical
           rank of :math:`A` when it is rank-deficient
    """
    assert a.shape[1] < a.shape[0], 'a must be tall'
    # QR decomposition
    if mode == 'complete':
        q, r = np.linalg.qr(a, mode='complete')
    else:
        q, r = np.linalg.qr(a)
    # check for rank
    rank = np.count_nonzero(np.abs(np.diag(r)) > epsilon)
    # return tuple
    if mode == 'complete':
        return rank, q[:, :rank], q[:, rank:]
    else:
        return rank, q[:, :rank]


def is_colinear(a: npt.NDArray, b: npt.NDArray, epsilon: float = 1e-8) -> bool:
    """
    Two vectors are colinear if their orthogonal projection is small

    :param a: first vector
    :param b: second vector
    :param epsilon: accuracy
    :return: ``True`` if colinear
    """
    return np.linalg.norm(a - (np.dot(a, b) / np.dot(b, b)) * b) \
        < epsilon * np.linalg.norm(a)


def norm(a: Union[npt.NDArray, scipy.sparse.csr_matrix]) -> float:
    """
    :param a: the array
    :return: the 2-norm of the array
    """
    return np.linalg.norm(a.data) if scipy.sparse.issparse(a) else np.linalg.norm(a)


class Colors(Enum):
    """
    Default colors
    """

    BLUE = (0, 0.4470, 0.7410)
    """
    """
    ORANGE = (0.8500, 0.3250, 0.0980)
    """
    """
    YELLOW = (0.9290, 0.6940, 0.1250)
    """
    """
    PURPLE = (0.4940, 0.1840, 0.5560)
    """
    """
    GREEN = (0.4660, 0.6740, 0.1880)
    """
    """
    LIGHT_BLUE = (0.3010, 0.7450, 0.9330)
    """
    """
    BROWN = (0.6350, 0.0780, 0.1840)
    """
    """



import unittest

import numpy as np

from tnsgrt.utils import rotation_2d, rotation_3d


class TestUtils(unittest.TestCase):

    def test_rotation_2d(self):

        v = np.array([1, 0])
        rotation = rotation_2d(0)
        np.testing.assert_array_equal(rotation @ v, v)

        rotation = rotation_2d(np.pi/2)
        np.testing.assert_almost_equal(rotation @ v, np.array([0, 1]))

        rotation = rotation_2d(np.pi)
        np.testing.assert_almost_equal(rotation @ v, np.array([-1, 0]))

        rotation = rotation_2d(np.pi/4)
        np.testing.assert_almost_equal(rotation @ v, np.array([1, 1])/np.sqrt(2))

    def test_rotation_3d(self):

        v = np.array([1, 0, 0])
        z = np.array([0, 0, 1])
        rotation = rotation_3d(0 * z)
        np.testing.assert_array_equal(rotation @ v, v)

        rotation = rotation_3d(np.pi/2 * z)
        np.testing.assert_almost_equal(rotation @ v, np.array([0, 1, 0]))

        rotation = rotation_3d(np.pi * z)
        np.testing.assert_almost_equal(rotation @ v, np.array([-1, 0, 0]))

        rotation = rotation_3d(np.pi/4 * z)
        np.testing.assert_almost_equal(rotation @ v, np.array([1, 1, 0])/np.sqrt(2))

        z = np.array([0, -1, 0])
        rotation = rotation_3d(0 * z)
        np.testing.assert_array_equal(rotation @ v, v)

        rotation = rotation_3d(np.pi/2 * z)
        np.testing.assert_almost_equal(rotation @ v, np.array([0, 0, 1]))

        rotation = rotation_3d(np.pi * z)
        np.testing.assert_almost_equal(rotation @ v, np.array([-1, 0, 0]))

        rotation = rotation_3d(np.pi/4 * z)
        np.testing.assert_almost_equal(rotation @ v, np.array([1, 0, 1])/np.sqrt(2))



from typing import Tuple, Optional

import numpy as np
import numpy.typing as npt
import matplotlib.pyplot as plt

from .plotter import Plotter
from tnsgrt.structure import Structure


class MatplotlibPlotter(Plotter):
    """
    Matplotlib based structure plotter

    :param plotter: :class:`tnsgrt.plotter.matplotlib.Matplotlib` object
    :param fig: :class:`matplotlib.figure` object
    :param ax: :class:`matplotlib.pyplot.axis` object
    """

    defaults = {
        'plot_nodes': True,
        'plot_members': True,
        'node_marker': 'o',
        'node_markersize': 4,
        'node_linewidth': 2,
        'node_linestyle': 'none',
        'node_facecolor': (0, 0, 0),
        'node_edgecolor': (1, 1, 1)
    }

    @staticmethod
    def plot_line(ax: plt.Axes, nodes: npt.NDArray[np.float_],
                  i: int, j: int, **kwargs) -> None:
        """
        Plot line connecting ``nodes[i]`` to ``nodes[j]``

        :param ax: :class:`matplotlib.pyplot.axis` object
        :param nodes: 3 x m array of nodes
        :param i: the index of the beginning of the line
        :param j: the index of the end of the line
        :param \**kwargs: additional keywords arguments passed to
                          ``matplotlib.axis.plot``
        """
        # draw lines
        x = np.hstack((nodes[0, i], nodes[0, j]))
        y = np.hstack((nodes[1, i], nodes[1, j]))
        z = np.hstack((nodes[2, i], nodes[2, j]))
        ax.plot(x, y, z, **kwargs)

    @staticmethod
    def plot_solid_cylinder(ax: plt.Axes, nodes: npt.NDArray[np.float_],
                            i: int, j: int, volume: float = 0., radius: float = 0.01,
                            n: int = 12, **kwargs) -> None:
        """
        Plot solid cylinder connecting ``nodes[i]`` to ``nodes[j]``

        :param ax: :class:`matplotlib.pyplot.axis` object
        :param nodes: 3 x m array of nodes
        :param i: the index of the beginning of the line
        :param j: the index of the end of the line
        :param volume: the cylinder volume
        :param radius: the cylinder radius
        :param n: the number of sides of the cylinder
        :param \**kwargs: additional keywords arguments passed to
                          ``matplotlib.axis.plot``
        """

        # cylinder nodes
        x, y, z = Plotter.cylinder(nodes[:, j], nodes[:, i], volume, radius, n)

        ax.plot_surface(x, y, z, **kwargs)

    def __init__(self, plotter: Optional['MatplotlibPlotter'] = None,
                 fig: Optional[plt.Figure] = None, ax: Optional[plt.Axes] = None):
        # call super
        super().__init__()

        # initialize plotter
        if plotter is not None:
            if ax is not None or fig is not None:
                raise 'fig and ax handles can not be given when plotter is given'
            self.fig, self.ax = plotter.get_handles()

        elif fig is None:
            if ax is not None:
                raise 'fig handle must be given when ax is given'
            # initialize figure and axis
            self.fig, self.ax = plt.subplots(subplot_kw={"projection": "3d"})

        elif ax is None:
            # initialize axis
            self.ax = fig.subplots(subplot_kw={"projection": "3d"})

        else:
            # assign fig and axis
            self.fig, self.ax = fig, ax

    def get_handles(self) -> Tuple[plt.Figure, plt.Axes]:
        """
        :return: tuple with matplotlib figure and axis
        """
        return self.fig, self.ax

    def view_init(self, elev=0, azim=0, roll=0) -> None:
        """
        Set view

        :param elev: elevation
        :param azim: azimuth
        :param roll: roll
        """
        self.ax.view_init(elev, azim, roll)

    def plot(self, *s: Structure, **kwargs) -> None:

        # loop if more than one is given
        if len(s) != 1:
            for si in s:
                self.plot(si, **kwargs)
            return
        else:
            s = s[0]

        # process options
        defaults = MatplotlibPlotter.defaults.copy()
        defaults.update(kwargs)

        # plot nodes
        nodes = s.nodes
        if defaults['plot_nodes']:
            self.ax.plot(nodes[0, :], nodes[1, :], nodes[2, :],
                         defaults['node_marker'],
                         markersize=defaults['node_markersize'],
                         linewidth=defaults['node_linewidth'],
                         linestyle=defaults['node_linestyle'],
                         markerfacecolor=defaults['node_facecolor'],
                         markeredgecolor=defaults['node_edgecolor'])

        # plot members
        members = s.members
        if defaults['plot_members']:
            for j in range(s.get_number_of_members()):
                if s.member_properties.loc[j, 'visible']:
                    if s.has_member_tag(j, 'string'):
                        # plot strings as lines
                        kwargs = s.get_member_properties(j, 'facecolor',
                                                         'linewidth').to_dict()
                        kwargs['color'] = kwargs['facecolor']
                        del kwargs['facecolor']
                        MatplotlibPlotter.plot_line(self.ax, nodes, members[0, j],
                                                    members[1, j], **kwargs)
                    else:
                        # plot others as solid elements
                        kwargs = s.get_member_properties(j, 'facecolor', 'edgecolor',
                                                         'volume', 'radius').to_dict()
                        MatplotlibPlotter.plot_solid_cylinder(self.ax, nodes,
                                                              members[0, j],
                                                              members[1, j], **kwargs)



import unittest
from typing import Optional

import numpy as np

from tnsgrt.prism import Prism
from tnsgrt.stiffness import NodeConstraint


class TestStiffness(unittest.TestCase):

    def test_stiffness(self):
        s = Prism(3)
        S, F, v = s.stiffness()
        d = S.eigs()[0]
        # 6 rigid body nodes
        self.assertEqual(np.count_nonzero(d < 1e-3), 6)
        # 6 rigid body nodes + 1 soft node
        self.assertEqual(np.count_nonzero(d < 1e6), 7)

    def test_rigid_body_1(self):
        s = Prism(3)
        # calculate stiffness
        S, F, v = s.stiffness(apply_rigid_body_constraint=True)
        # calculate eigenvalues
        d = S.eigs(k=6)[0]
        # no rigid body nodes
        self.assertEqual(np.count_nonzero(d < 1e-3), 0)
        # 1 soft node
        self.assertEqual(np.count_nonzero(d < 1e6), 1)

    def test_eigenmodes(self):
        s = Prism(3)
        # calculate stiffness
        S, F, v = s.stiffness(apply_rigid_body_constraint=True)
        # eigenmodes
        d = S.modes(k=6)[0]
        # no rigid body nodes
        self.assertEqual(np.count_nonzero(d < 1e-3), 0)
        # 1 soft node
        self.assertEqual(np.count_nonzero(d < 1e2), 1)

    def test_node_constraint(self):

        c = NodeConstraint()
        self.assertEqual(c.dof, 0)
        np.testing.assert_allclose(c.normal, np.eye(3))
        self.assertEqual(c.basis, None)

        c = NodeConstraint(np.array([[1, 0, 0]]))
        self.assertEqual(c.dof, 2)
        np.testing.assert_allclose(c.normal, np.array([[1, 0, 0]]))
        np.testing.assert_allclose(c.basis, [[0, 0], [1, 0], [0, 1]])

        c = NodeConstraint(np.array([[1, 0, 0], [0, 1, 0]]))
        self.assertEqual(c.dof, 1)
        np.testing.assert_allclose(c.normal, np.array([[1, 0, 0], [0, 1, 0]]))
        np.testing.assert_allclose(c.basis, [[0], [0], [1]])

        c = NodeConstraint(np.array([[1, 1, 0]]))
        self.assertEqual(c.dof, 2)
        np.testing.assert_allclose(c.normal,
                                   np.array([[-np.sqrt(2) / 2, -np.sqrt(2) / 2, 0]]))
        np.testing.assert_allclose(c.basis,
                                   [[-np.sqrt(2) / 2, 0], [np.sqrt(2) / 2, 0], [0, 1]])

        nodes = np.array([[1, 0], [0, 1], [0, 0]])
        constraints = [None]*2
        constraints[0] = NodeConstraint(np.array(([[1, 1, 0]])))
        constraints[1] = NodeConstraint()

        R, T = NodeConstraint.node_constraint(nodes, constraints, storage='dense')
        self.assertEqual(R.shape, (4, 6))
        self.assertEqual(T.shape, (6, 2))

        Rs, Ts = NodeConstraint.node_constraint(nodes, constraints)
        self.assertEqual(Rs.shape, (4, 6))
        self.assertEqual(Ts.shape, (6, 2))

        np.testing.assert_equal(R, Rs.toarray())
        np.testing.assert_equal(T, Ts.toarray())

        self.assertTrue(np.linalg.norm((Rs @ Ts).toarray()) < 1e-6)
        self.assertTrue(np.linalg.norm(Rs @ Rs.transpose() - np.eye(4)) < 1e-6)
        self.assertTrue(np.linalg.norm(Ts.transpose() @ Ts - np.eye(2)) < 1e-6)

        nodes = np.array([[1, 0, 1], [0, 1, 1], [0, 0, 1]])
        constraints: list[Optional[NodeConstraint]] = [None]*3
        constraints[0] = NodeConstraint(np.array(([[1, 1, 0]])))
        constraints[2] = NodeConstraint()

        R, T = NodeConstraint.node_constraint(nodes, constraints, storage='dense')
        self.assertEqual(R.shape, (4, 9))
        self.assertEqual(T.shape, (9, 5))

        Rs, Ts = NodeConstraint.node_constraint(nodes, constraints)
        self.assertEqual(Rs.shape, (4, 9))
        self.assertEqual(Ts.shape, (9, 5))

        np.testing.assert_equal(R, Rs.toarray())
        np.testing.assert_equal(T, Ts.toarray())

        self.assertTrue(np.linalg.norm((Rs @ Ts).toarray()) < 1e-6)
        self.assertTrue(np.linalg.norm(Rs @ Rs.transpose() - np.eye(4)) < 1e-6)
        self.assertTrue(np.linalg.norm(Ts.transpose() @ Ts - np.eye(5)) < 1e-6)

        nodes = np.array([[1, 0, 0, 0, 1], [0, 1, 0, 1, 1], [0, 0, 2, 1, 3]])
        constraints: list[Optional[NodeConstraint]] = [None]*5
        constraints[0] = NodeConstraint(np.array(([[1, 1, 0]])))
        constraints[2] = NodeConstraint()

        R, T = NodeConstraint.node_constraint(nodes, constraints, storage='dense')
        self.assertEqual(R.shape, (4, 15))
        self.assertEqual(T.shape, (15, 11))

        Rs, Ts = NodeConstraint.node_constraint(nodes, constraints)
        self.assertEqual(Rs.shape, (4, 15))
        self.assertEqual(Ts.shape, (15, 11))

        np.testing.assert_equal(R, Rs.toarray())
        np.testing.assert_equal(T, Ts.toarray())

        self.assertTrue(np.linalg.norm((Rs @ Ts).toarray()) < 1e-6)
        self.assertTrue(np.linalg.norm(Rs @ Rs.transpose() - np.eye(4)) < 1e-6)
        self.assertTrue(np.linalg.norm(Ts.transpose() @ Ts - np.eye(11)) < 1e-6)

    def test_rigid_body_2(self):

        nodes = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]]).transpose()
        constraints = NodeConstraint.rigid_body_three_point_constraint(nodes)
        self.assertEqual(constraints[0].dof, 0)
        self.assertEqual(constraints[1].dof, 1)
        np.testing.assert_allclose(constraints[1].normal,
                                   np.array([[0, 0, -1], [0, 1, 0]]))
        self.assertEqual(constraints[2].dof, 2)
        np.testing.assert_allclose(constraints[2].normal, np.array([[0, 0, -1]]))

        R, T = NodeConstraint.node_constraint(nodes, constraints, storage='dense')
        self.assertEqual(R.shape, (6, 9))
        self.assertEqual(T.shape, (9, 3))

        Rs, Ts = NodeConstraint.node_constraint(nodes, constraints)
        self.assertEqual(Rs.shape, (6, 9))
        self.assertEqual(Ts.shape, (9, 3))

        np.testing.assert_equal(R, Rs.toarray())
        np.testing.assert_equal(T, Ts.toarray())

    def test_rigid_body_3(self):

        # create 3 prism
        s = Prism(3)
        # set rigid body constraint on 3 bottom nodes
        s.node_properties.loc[:2, 'constraint'] = \
            NodeConstraint.rigid_body_three_point_constraint(s.nodes[:, :3])
        # calculate stiffness
        S, F, v = s.stiffness()
        # calculate eigenvalues
        d = S.eigs(k=6)[0]
        # no rigid body nodes
        self.assertEqual(np.count_nonzero(d < 1e-3), 0)
        # 1 soft node
        self.assertEqual(np.count_nonzero(d < 1e6), 1)


if __name__ == '__main__':
    unittest.main()



import warnings

import numpy as np

from .structure import Structure


class Michell(Structure):
    """
    Constructs a circular Michell truss with ``n`` sides and ``q`` layers

    :param n: the number of sides
    :param beta: the angle between the radius and the outermost bar
    :param q: the number of layers
    :param radius: the radius of the structure
    :param \\**kwargs: See below

    :Keyword Arguments:
        * **spiral** (``bool=True``) --
          if ``True`` add *spiral members*
        * **radial** (``bool=True``) --
          if ``True`` add *radial members*
        * **outer** (``bool=True``) --
          if ``True`` add *outer members*
        * **inner** (``bool=True``) --
          if ``True`` add *inner members*
        * **center** (``bool=True``) --
          if ``True`` add *center members*

    **Notes:**

    1. If :math:`\\rho` is equal to ``radius``, then the radii as a function of the
       layer number `i` is

      .. math::
           r(i) = \\rho \\, a^i, \\qquad a = \\frac{\\beta}{\\sin(\\beta + \\pi)}

    2. The radii are convergent only if :math:`a < 1`, that is

       .. math::
          \\beta < \\frac{\\pi}{2} - \\frac{\\pi}{2 n}

       The default :math:`\\beta = \\pi/4` leads to convergent designs for all
       :math:`n > 2`.
    3. Additional keyword arguments are passed to :class:`tnsgrt.structure.Structure`
    """

    def __init__(self,
                 n: int = 6, beta: float = np.pi/4, q: int = 4, radius: float = 1,
                 **kwargs):
        # other options
        options = {
            'spiral': True,
            'radial': True,
            'outer': True,
            'inner': True,
            'center': True,
        }
        options.update(kwargs)

        # remove used options
        kwargs = {k: v for k, v in kwargs.items()
                  if k not in ['spiral', 'radial', 'outer', 'inner', 'center']}

        # wedge angle
        phi = np.pi / n

        # radius ratio
        a = np.sin(beta) / np.sin(beta + phi)

        # warn if not convergent
        if phi + 2 * beta > np.pi:
            warnings.warn('radius are not convergent')

        # nodes
        nodes = np.zeros((3*n, q))
        for i in range(n):
            for k in range(q):
                nodes[3*i:3*(i+1), k] = [radius * (a ** k) * np.cos((2*i - k) * phi),
                                         radius * (a ** k) * np.sin((2*i - k) * phi),
                                         0]

        def node_index(i, j):
            return n * j + i

        members = np.zeros((2, 0))
        # add spiral members
        if options['spiral']:
            michell_members = np.zeros((2*n, 2*(q-1)))
            for i in range(n):
                for j in range(q-1):
                    michell_members[2*i:2*(i+1), j] = \
                        [node_index(i, j), node_index(np.mod(i+1, n), np.mod(j+1, q))]
                    michell_members[2*i:2*(i+1), q-1+j] = \
                        [node_index(i, j), node_index(i, np.mod(j+1, q))]

            # reshape members
            members = np.hstack((members,
                                 michell_members.reshape((2, n*2*(q-1)), order='F')))

        # add radial
        if options['radial'] and q > 2:
            m = q-2
            radial_members = np.zeros((2*n, m))
            for i in range(n):
                for j in range(m):
                    radial_members[2*i:2*(i+1), j] = \
                        [node_index(i, j), node_index(np.mod(i+1, n), np.mod(j+2, q))]

            # reshape members
            members = np.hstack((members, radial_members.reshape((2, n*m), order='F')))

        # add outer
        if options['outer']:
            outer_members = np.zeros((2*n, 1))
            for i in range(n):
                outer_members[2*i:2*(i+1), 0] = \
                    [node_index(i, 0), node_index(np.mod(i+1, n), 0)]

            # reshape members
            members = np.hstack((members, outer_members.reshape((2, n), order='F')))

        # add inner
        if options['inner']:
            inner_members = np.zeros((2*n, 1))
            for i in range(n):
                inner_members[2*i:2*(i+1), 0] = \
                    [node_index(i, q-1), node_index(np.mod(i+1, n), q-1)]

            # reshape members
            members = np.hstack((members, inner_members.reshape((2, n), order='F')))

        # reshape nodes
        nodes = nodes.reshape((3, n * q), order='F')

        # add center
        if options['center']:
            nodes = np.hstack((nodes, np.zeros((3, 1))))
            center_node_index = nodes.shape[1] - 1
            center_members = np.zeros((2*n, 1))
            for i in range(n):
                center_members[2*i:2*(i+1), 0] = [node_index(i, q-1), center_node_index]

            # reshape members
            members = np.hstack((members, center_members.reshape((2, n), order='F')))

        # print(nodes)
        # print(members)

        # call super
        super().__init__(nodes, members, **kwargs)



import warnings
from typing import Optional, Tuple, Sequence, Union
from typing_extensions import Literal

import numpy as np
import numpy.typing as npt
import scipy

from .utils import orthogonalize, norm


class NodeConstraint:
    """
    Object representing a node constraint

    :param constraint: the constraint coefficient array
    :param epsilon: precision used to assess numerical rank

    **Notes:**

    1. If :math:`x` is the node coordinate and :math:`A` is the constraint coefficient
       then

       .. math::
           A x = 0

       then the equivalent constraint and its null space

       .. math::
           R \\, x = 0, \\quad x = T \\, y, \\quad R R^T = I, \\quad T^T T = I

       is constructed using :func:`tnsgrt.utils.orthogonalize`
    """

    def __init__(self, constraint: Optional[npt.NDArray] = None, epsilon: float = 1e-8):
        if constraint is None:
            self.dof = 0
            self.basis = None
            self.normal = np.eye(3)
        else:
            assert constraint.ndim == 2 and constraint.shape[1] == 3 and \
                   constraint.shape[0] < 3, 'normal must be a m x 3, m < 3, array'
            # first orthogonalize
            rank, r, t = orthogonalize(constraint.transpose(),
                                       mode='complete', epsilon=epsilon)
            self.dof = 3 - rank
            if self.dof == 0:
                self.basis = None
                self.normal = np.eye(3)
            else:
                self.normal = r.transpose()
                self.basis = t

    @staticmethod
    def node_constraint(nodes: npt.NDArray, constraints: Sequence['NodeConstraint'],
                        storage: Literal['sparse', 'dense'] = 'sparse') \
            -> Union[Tuple[npt.NDArray, npt.NDArray],
                     Tuple[scipy.sparse.csr_array, scipy.sparse.csr_array]]:
        """
        Construct the constraint associated with the given nodes and node constraints

        The resulting tuple is compatible with
        :meth:`~tnsgrt.stiffness.Stiffness.apply_constraint`

        :param nodes: 3 x n array of nodes
        :param constraints: list with n constraints
        :param storage: if ``sparse``, returns sparse arrays
        :return: tuple with the constraint matrix, and its null space
        """
        assert nodes.shape[0] == 3, 'nodes must be a 3 x m array'
        assert nodes.shape[1] == len(constraints), \
            'number of columns of nodes must be equal to to the number of constraints'

        # calculate degrees of freedom
        m = nodes.shape[1]
        noc = sum(3 - c.dof for c in constraints if c is not None)
        dof = 3*m - noc

        if storage == 'sparse':
            row_col_r = np.zeros((2, 3*noc))
            data_r = np.zeros((3*noc))
            row_col_t = np.zeros((2, 3*dof))
            data_t = np.zeros((3*dof))
            jj = kk = 0
            for i, c in enumerate(constraints):
                if c is None:
                    # add identity to basis
                    dofj = 3
                    row_col_t[0, 3*kk:3*kk+3*dofj] = \
                        np.kron(np.arange(3*i, 3*(i+1)), np.ones((3,)))
                    row_col_t[1, 3*kk:3*kk+3*dofj] = \
                        np.kron(np.arange(kk, kk+dofj), np.ones((3,)))
                    data_t[3*kk:3*kk+3*dofj] = np.eye(dofj).flatten(order='C')
                    kk += dofj
                else:
                    # add to normal
                    nocj = 3-c.dof
                    row_col_r[0, 3*jj:3*jj+3*nocj] = \
                        np.kron(np.arange(jj, jj+nocj), np.ones((3,)))      # row
                    row_col_r[1, 3*jj:3*jj+3*nocj] = \
                        np.kron(np.ones((nocj,)), np.arange(3*i, 3*(i+1)))  # col
                    data_r[3*jj:3*jj+3*nocj] = c.normal.flatten(order='C')
                    jj += nocj
                    if c.dof:
                        # add to basis
                        dofj = c.dof
                        row_col_t[0, 3*kk:3*kk+3*dofj] = \
                            np.kron(np.arange(3*i, 3*(i+1)), np.ones((dofj,)))  # row
                        row_col_t[1, 3*kk:3*kk+3*dofj] = \
                            np.kron(np.ones((3,)), np.arange(kk, kk+dofj))      # col
                        data_t[3*kk:3*kk+3*dofj] = c.basis.flatten(order='C')
                        kk += dofj
            R = scipy.sparse.coo_matrix((data_r, row_col_r),
                                        shape=(noc, 3*m)).tocsr()
            T = scipy.sparse.coo_matrix((data_t, row_col_t),
                                        shape=(3*m, dof)).tocsr()
        else:
            R = np.zeros((noc, 3*m))
            T = np.zeros((3*m, dof))
            jj = kk = 0
            for i, c in enumerate(constraints):
                if c is None:
                    # add identity to basis
                    T[3*i:3*(i+1), kk:kk+3] = np.eye(3)
                    kk += 3
                else:
                    # add to constraint
                    nocj = 3-c.dof
                    R[jj:jj+nocj, 3*i:3*(i+1)] = c.normal
                    jj += nocj
                    if c.dof:
                        # add to basis
                        T[3*i:3*(i+1), kk:kk+c.dof] = c.basis
                        kk += c.dof

        return R, T

    @staticmethod
    def rigid_body_three_point_constraint(nodes: npt.NDArray, epsilon: float = 1e-8) \
            -> Tuple['NodeConstraint', 'NodeConstraint', 'NodeConstraint']:
        """
        Return three constraints on three given nodes to prevent rigid motion.

        :param nodes: 3 x 3 array of nodes (column oriented)
        :param epsilon: accuracy to assess co-linearity
        :return: tuple with 3 node constraints

        **Notes:**

        1. The constraints are as follows:
           - Node 0 is fixed
           - Node 1 is allowed to move in the line 0-1
           - Node 2 is allowed to move in the plane 0-1-2
        """
        assert nodes.shape[0] == 3 and nodes.shape[1] == 3, '3 nodes should be provided'

        # determine vector normal to plane
        v10 = (nodes[:, 1] - nodes[:, 0]).flatten()
        v20 = (nodes[:, 2] - nodes[:, 0]).flatten()
        v012p = np.cross(v10, v20)

        if np.linalg.norm(v012p) < epsilon:
            raise Exception('the three given points do not form a plane')

        # determine normal to vp and v10
        v10p = np.cross(v10, v012p)

        # constraints are:
        # - node 0 to be fixed,
        # - node 1 to be in line from 0 to 1,
        # - node 2 to be in the plane
        return NodeConstraint(), NodeConstraint(np.vstack((v012p, v10p))), \
            NodeConstraint(v012p.reshape((1, 3)))

    @staticmethod
    def rigid_body_constraint(nodes: npt.NDArray, epsilon: float = 1e-8) -> \
            Tuple[npt.NDArray, npt.NDArray]:
        """
        Construct rigid-body constraint associated with the given nodes

        The resulting tuple is compatible with
        :meth:`~tnsgrt.stiffness.Stiffness.apply_constraint`

        :param nodes: 3 x n array of nodes
        :param epsilon: precision used to assess numerical rank
        :return: tuple with the constraint matrix, and its null space
        """

        assert len(nodes.shape) == 2 and nodes.shape[0] == 3, \
            'nodes must be a 3 x m matrix'

        number_of_nodes = nodes.shape[1]

        # rigid translation
        t1 = np.zeros((3, number_of_nodes), order='F')
        t1[0, :] = 1 / np.sqrt(number_of_nodes)

        t2 = np.zeros((3, number_of_nodes), order='F')
        t2[1, :] = 1 / np.sqrt(number_of_nodes)

        t3 = np.zeros((3, number_of_nodes), order='F')
        t3[2, :] = 1 / np.sqrt(number_of_nodes)

        # rigid rotations
        r1 = (np.array([[0, 0, 0], [0, 0, -1], [0, 1, 0]]) @ nodes).ravel(order='F')
        r2 = (np.array([[0, 0, 1], [0, 0, 0], [-1, 0, 0]]) @ nodes).ravel(order='F')
        r3 = (np.array([[0, -1, 0], [1, 0, 0], [0, 0, 0]]) @ nodes).ravel(order='F')

        r = np.vstack((t1.ravel(order='F'), t2.ravel(order='F'), t3.ravel(order='F'),
                       r1 / np.linalg.norm(r1),
                       r2 / np.linalg.norm(r2),
                       r3 / np.linalg.norm(r3)))

        rank, r, t = orthogonalize(r.transpose(), epsilon=epsilon, mode='complete')

        return r.transpose(), t

    @staticmethod
    def planar_constraint(nodes: npt.NDArray, normal: Optional[npt.NDArray] = None,
                          epsilon: float = 1e-8, storage='sparse'):
        """
        Construct the planar constraint associated with the given nodes and normal
        vector

        The resulting tuple is compatible with
        :meth:`~tnsgrt.stiffness.Stiffness.apply_constraint`

        :param nodes: 3 x n array of nodes
        :param normal: list with n constraints
        :param storage: if ``sparse``, returns sparse arrays
        :return: tuple with the constraint matrix, and its null space
        """
        assert nodes.shape[0] == 3, 'nodes must be a 3 x m array'
        if normal is None:
            normal = np.array([[0], [0], [1]])
        elif normal.ndim == 1 and normal.shape[0] == 3:
            normal = normal.reshape((3, 1))
        elif normal.ndim == 2 and normal.shape[0] == 1 and normal.shape[1] == 3:
            normal = normal.transpose()
        assert normal.ndim == 2 and normal.shape[0] == 3 and normal.shape[1] == 1, \
            'normal must be a 3D vector'

        # calculate normal and basis
        _, normal, basis = orthogonalize(normal, mode='complete', epsilon=epsilon)

        # calculate degrees of freedom
        n = nodes.shape[1]
        noc = n
        dof = 2*n

        if storage == 'sparse':
            row_col_r = np.zeros((2, 3*noc))
            data_r = np.zeros((3*noc))
            row_col_t = np.zeros((2, 3*dof))
            data_t = np.zeros((3*dof))
            jj = kk = 0
            # flatten normal and basis
            normal = normal.flatten(order='C')
            basis = basis.flatten(order='C')
            nocj, dofj = 1, 2
            for i in range(n):
                # add to constraint
                row_col_r[0, 3*jj:3*jj+3*nocj] = \
                    np.kron(np.arange(jj, jj+nocj), np.ones((3,)))      # row
                row_col_r[1, 3*jj:3*jj+3*nocj] = \
                    np.kron(np.ones((nocj,)), np.arange(3*i, 3*(i+1)))  # col
                data_r[3*jj:3*jj+3*nocj] = normal
                jj += nocj
                # add to basis
                row_col_t[0, 3*kk:3*kk+3*dofj] = \
                    np.kron(np.arange(3*i, 3*(i+1)), np.ones((dofj,)))  # row
                row_col_t[1, 3*kk:3*kk+3*dofj] = \
                    np.kron(np.ones((3,)), np.arange(kk, kk+dofj))      # col
                data_t[3*kk:3*kk+3*dofj] = basis
                kk += dofj
            R = scipy.sparse.coo_matrix((data_r, row_col_r),
                                        shape=(noc, 3*n)).tocsr()
            T = scipy.sparse.coo_matrix((data_t, row_col_t),
                                        shape=(3*n, dof)).tocsr()
        else:
            R = np.zeros((noc, 3*n))
            T = np.zeros((3*n, dof))
            jj = kk = 0
            nocj, dofj = 1, 2
            normal = normal.flatten()
            for i in range(n):
                # add to constraint
                R[jj:jj+nocj, 3*i:3*(i+1)] = normal
                jj += nocj
                # add to basis
                T[3*i:3*(i+1), kk:kk+dofj] = basis
                kk += dofj

        return R, T


class Stiffness:
    """
    Model for a constrained mechanical system consisting of

    - :math:`K`: the stiffness matrix
    - :math:`M`: the mass matrix
    - :math:`R`: a matrix representing node displacement constraints
    - :math:`T`: a matrix representing allowed node displacements

    such that

    .. math::
        V = \\frac{1}{2} y^T K y + \\frac{1}{2} \\ddot{y}^T M \\ddot{y},
        \\qquad R \\, x = 0, \\qquad x = T y

    Constructor parameters:

    :param K: the stiffness matrix
    :param M: the mass matrix

    **Notes:**

    1. The stiffness and mass matrices are stored in the reduced coordinates :math:`y`

    2. In global coordinates

       .. math::
           V = \\frac{1}{2} x^T K_x x + \\frac{1}{2} \\ddot{x}^T M_x \\ddot{x}

       in which

       .. math::
           K_x = T^T K T, \\qquad M_x = T^T M T

    3. Use :meth:`tnsgrt.stiffness.Stiffness.apply_constraint` to apply constraints
       to the model

    4. Node constraints are enforced to be orthogonal so that

       .. math::
           R \\, x = 0, \\quad x = T y, \\qquad R R^T = I,
           \\quad T^T T = I, \\quad R \\, T = 0
    """

    def __init__(self,
                 K: Union[npt.NDArray, scipy.sparse.csr_matrix],
                 M: Optional[Union[npt.NDArray, scipy.sparse.csr_matrix]] = None):

        self.K = K
        self.M = M
        self.T: Optional[Union[npt.NDArray, scipy.sparse.csr_matrix]] = None
        self.R: Optional[Union[npt.NDArray, scipy.sparse.csr_matrix]] = None

        assert K.ndim == 2 and K.shape[0] == K.shape[1], 'K must be a square matrix'

        if M is not None:
            assert M.ndim == 2 and M.shape[0] == M.shape[1] \
                   and M.shape[0] == K.shape[0], \
                'M must be a square matrix of same size as K'

    def apply_constraint(self,
                         R: Union[npt.NDArray, scipy.sparse.csr_matrix],
                         T: Optional[Union[npt.NDArray, \
                                           scipy.sparse.csr_matrix]] = None,
                         local: bool = True, epsilon: float = 1e-8):
        """
        Apply the constraint

        .. math::
            R \\, y = 0, \\qquad y = T z

        to the current or the global ``Stiffness`` object coordinate

        :param R: the constraint coefficient matrix
        :param T: the allowed node displacements; if ``None`` constraint is normalized
        :param local: if ``True`` the constraint is applied
        :param epsilon: precision used to assess numerical rank
        :return:

        **Notes:**

        1. If ``local = True`` and the current constraints are

           .. math::
               R_y \\, x = 0, \\quad x = T_y \\, y, \\qquad R^{}_y R_y^T = I,
               \\quad T_y^T T^{}_y = I, \\quad R_y T_y = 0

           then after applying the new constraints

           .. math::
               R \\, y = R \\, T_y^T x = R_z x = 0, \\qquad x = T_y \\,
               y = T_y T z = T_z z,

           in which

           .. math::
               R_z = R \\, T_y^T, \\qquad T_z = T_y T

           and

           .. math::
               V = \\frac{1}{2} z^T K_z \\, z + \\frac{1}{2} \\ddot{z}^T M_z \\,
               \\ddot{z}, \\qquad R_z x = 0, \\quad x = R_z \\, z,

           in which

           .. math::
               K_z = T^T K T, \\qquad M_z = T^T M T

        2. If ``local = False`` and the constraints are interpreted as

           .. math::
               R \\, x = 0, \\quad x = T \\, z

           which is first converted to the local constraint

           .. math::
               R \\, x = R \\, T_y y = \\tilde{R} y = 0, \\qquad \\tilde{R} = R \\, T_y,

           before applying
        """

        if local or self.T is None:
            # local constraint
            if T is None:
                # normalize before applying
                rank, r, t = orthogonalize(R.transpose(),
                                           mode='complete', epsilon=epsilon)
                r = r.transpose()
            else:
                # test orthogonality
                assert norm(R @ T) < epsilon, 'R and T must be orthogonal'
                assert norm(R @ R.transpose() - scipy.sparse.eye(R.shape[0])) \
                       < epsilon, 'R must be unitary'
                assert norm(T.transpose() @ T - scipy.sparse.eye(T.shape[1])) \
                       < epsilon, 'T must be unitary'
                r, t = R, T

            # apply constraint

            if scipy.sparse.issparse(self.K):
                # make sure it is sparse
                if scipy.sparse.issparse(r):
                    r = scipy.sparse.csr_matrix(r)
                if scipy.sparse.issparse(t):
                    t = scipy.sparse.csr_matrix(t)

            # project stiffness
            self.K = t.transpose() @ self.K @ t

            # symmetrize
            self.K += self.K.transpose()
            self.K /= 2

            if self.M is not None:
                # project mass
                self.M = t.transpose() @ self.M @ t
                self.M += self.M.transpose()
                self.M /= 2

            if self.R is None:
                self.R = r
            else:
                if self.T is None:
                    self.R = r
                else:
                    self.R = r @ self.T.transpose()

            if self.T is None:
                self.T = t
            else:
                self.T = self.T @ t

        else:
            # not local and self.T is not None
            if T is not None:
                warnings.warn("allowed displacements parameter 'T' ignored when "
                              "'local=False'")
            self.apply_constraint(R @ self.T, epsilon=epsilon)

    def displacements(self, f: npt.NDArray):
        """
        Calculate displacements due to the application of a global force

        :param f: 3 x m array of forces
        :return: the displacements

        **Notes:**

        1. The displacements are calculated in global coordinates

           .. math::
               x = T K^{-1} T^T f
        """
        m = self.K.shape[0] if self.T is None else self.T.shape[0]
        assert f.shape[0] == 3 and f.shape[1] == m/3, 'f must be a 3 x m array'
        x = self.T @ np.linalg.solve(self.K,
                                     self.T.transpose() @ f.flatten(order='F')) \
            if self.T is not None else \
            np.linalg.solve(self.K, f.flatten(order='F'))
        return x.reshape((3, int(m/3)), order='F')

    def eigs(self, k: int = 12,
             which: Literal['LM', 'SM', 'LR', 'SR', 'LI', 'SI'] = 'SM'):
        """
        Compute the eigenvalues and eigenvectors of the stiffness matrix

        If the stiffness matrix is stored as a sparse array, return only the first k
        eigenvalue/eigenvector pairs

        :param k: the number of eigenvalues
        :param which: which eigenvalues to compute; see
                      :func:`scipy.sparse.linalg.eigsh` for details
        :return: tuple with eigenvalues, and eigenvectors

        **Notes:**

        1. The eigenvalues and eigenvectors are calculated by solving the eigenvalue
           problem

           .. math::
               K y = \\lambda \\, y, \\qquad x = T \\, y
        """

        if scipy.sparse.issparse(self.K):
            # calculate sparse eigenvalues
            d, v = scipy.sparse.linalg.eigsh(self.K, k=k, which=which)

        else:
            # calculate dense eigenvalues
            d, v = np.linalg.eigh(self.K)

        # sort
        ind = np.argsort(d)

        # project eigenvectors
        if self.T is not None:
            v = self.T @ v

        return d[ind], v[:, ind]

    def modes(self, k: int = 12, units='Hz', which: Literal['LM', 'SM'] = 'SM'):
        """
        Compute the natural frequencies [rad/s] and mode vectors

        :param k: the number of eigenvalues
        :param which: which eigenvalues to compute; see
                      :func:`scipy.sparse.linalg.eigsh` for details
        :param units: return frequencies in Hz if ``units = 'Hz'``
        :return: tuple with eigenvalues, and eigenvectors

        **Notes:**

        1. The natural frequencies and mode vectors are calculated by solving the
           generalized eigenvalue problem

           .. math::
               K y = \\omega^2 M y, \\qquad x = T \\, y
        """

        assert self.M is not None, 'modes cannot be calculated without mass information'

        if scipy.sparse.issparse(self.K):
            # calculate sparse eigenvalues
            d, v = scipy.sparse.linalg.eigsh(self.K, k=k, M=self.M, which=which)
        else:
            # calculate dense eigenvalues
            d, v = scipy.linalg.eigh(self.K, self.M)

        # calculate frequencies
        d = np.sqrt(np.abs(d))
        if units == 'Hz':
            d /= 2*np.pi

        # sort
        ind = np.argsort(d)

        # project eigenvectors
        if self.T is not None:
            v = self.T @ v

        return d[ind], v[:, ind]




# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
import pathlib


# sys.path.insert(0, os.path.abspath('../src'))
path = pathlib.Path(__file__).resolve() / '..' / '..' / 'src'
sys.path.insert(0, os.path.abspath(path))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'tnsgrt'
copyright = '2023, Mauricio de Oliveira'
author = 'Mauricio de Oliveira'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ['sphinx.ext.autodoc', 'sphinx.ext.mathjax']

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']
html_theme_options = {
    "navigation_depth": 4,
    "includehidden": True
}


import cvxpy as cvx
import numpy as np


def qp(n, m, Q, c, A, blo, bup, xlo, xup, **kwargs):
    """
    Minimize 1/2 x^T Q x + 1/2 f^T D f + c^T x
      s.t.   blo <= A x <= bup
             xlo <= x <= xup
             f = F^T x
    kwargs:
     - F and D
    """

    # variable
    x = cvx.Variable(n)

    # factored cost
    F = kwargs.pop('F', None)
    D = kwargs.pop('D', None)

    # objective function
    objective = 0
    if Q is not None:
        objective += (1 / 2) * cvx.quad_form(x, Q)
    if F is not None and D is not None:
        objective += (1 / 2) * cvx.quad_form(F.T @ x, D)
    if c is not None:
        objective += c.T @ x
    objective = cvx.Minimize(objective)

    # constraints
    constraints = []

    # variable upper bounds
    if xup is not None:
        isconstrained = np.logical_not(np.isinf(xup))
        constraints.append(x[isconstrained] <= xup[isconstrained])

    # variable lower bounds
    if xlo is not None:
        isconstrained = np.logical_not(np.isneginf(xlo))
        constraints.append(x[isconstrained] >= xlo[isconstrained])

    # constraints
    if m > 0:
        # upper and lower bounds
        if bup is not None and blo is not None:
            # equality constraints?
            isequal = blo == bup
            if np.any(isequal):
                constraints.append(A[isequal, :] @ x == bup[isequal])

            # upper bound constraints?
            isconstrained = np.logical_not(np.logical_or(np.isinf(bup), isequal))
            if np.any(isconstrained):
                constraints.append(A[isconstrained, :] @ x <= bup[isconstrained])

            isconstrained = np.logical_not(np.logical_or(np.isneginf(blo), isequal))
            if np.any(isconstrained):
                constraints.append(A[isconstrained, :] @ x >= blo[isconstrained])

        else:

            # a upper bounds
            if bup is not None:
                isconstrained = np.logical_not(np.isinf(bup))
                constraints.append(A[isconstrained, :] @ x <= bup[isconstrained])

            # a lower bounds
            if blo is not None:
                isconstrained = np.logical_not(np.isneginf(blo))
                constraints.append(A[isconstrained, :] @ x >= blo[isconstrained])

    # problem
    problem = cvx.Problem(objective, constraints)

    try:

        # solve
        defaults = {
            # 'solver': cvx.OSQP,
            # 'eps_abs': 1e-8
        }
        defaults.update(**kwargs)
        problem.solve(**defaults)

    except cvx.error.SolverError:

        # solve using default solver
        problem.solve(eps_abs=1e-8, **kwargs)

    return objective.value, x.value, problem.status


def lp(n, m, c, A, blo, bup, xlo, xup, **kwargs):
    """
    Minimize c^T x
      s.t.   blo <= A x <= bup
             xlo <= x <= xup
    """
    return qp(n, m, None, c, A, blo, bup, xlo, xup, **kwargs)



from typing import Tuple

import numpy as np
import numpy.typing as npt
from scipy.spatial.transform import Rotation

from tnsgrt.structure import Structure


class Plotter:
    """
    Base class for structure plotters
    """

    def plot(self, *s: Structure, **kwargs) -> None:
        """
        Plot structure

        :param \\*s: the structure or sequence of structures
        :param \\**kwargs: additional keyword arguments
        """
        pass

    @staticmethod
    def unit_cylinder(n: int = 10, radius: float = 1, height: float = 1) \
            -> tuple[npt.NDArray, npt.NDArray, npt.NDArray]:
        """
        Return xyz grid points for solid cylinder centered at the origin

        :param n: number of sides
        :param radius: radius of cylinder
        :param height: height of cylinder
        :return: tuple with x, y, and z points
        """
        z = np.linspace(0, height, 2)
        theta = np.linspace(0, 2 * np.pi, n)
        theta_grid, z_grid = np.meshgrid(theta, z)
        x_grid = radius * np.cos(theta_grid)
        y_grid = radius * np.sin(theta_grid)
        return x_grid, y_grid, z_grid

    @staticmethod
    def cylinder(node1: npt.NDArray, node2: npt.NDArray,
                 volume: float = 0., radius: float = 0.01, n: int = 12) \
            -> Tuple[npt.NDArray, npt.NDArray, npt.NDArray]:
        """
        Return xyz grid points for cylinder with center aligned with the vector defined
        by ``node1`` and ``node2``

        :param node1: center of the base of the cylinder
        :param node2: center of the top of the cylinder
        :param volume: volume of the cylinder, used to set the radius if positive
        :param radius: radius of the cylinder, ignore if volume is positive
        :param n: number of sides
        :return: tuple with x, y, and z points
        """
        # rod vector length
        rod = node2 - node1
        length = np.linalg.norm(rod)
        if volume > 0.:
            radius = np.sqrt((volume / length) / np.pi)

        # draw unit cylinder
        x, y, z = Plotter.unit_cylinder(n, radius, length)

        # rotation vector
        rotation_vector = [-rod[1], rod[0], 0]

        # rotation angle
        angle = np.arccos(rod[2] / length)
        if angle > 1e-4:

            norm = np.linalg.norm(rotation_vector)
            if norm < 1e-6:
                rotation_vector = np.array([1, 0, 0])
            else:
                rotation_vector /= norm

            rotation_matrix = Rotation.from_rotvec(angle * rotation_vector)

            # rotate
            for i in range(x.shape[0]):
                xyz = np.vstack((x[i, :], y[i, :], z[i, :]))
                xyz = rotation_matrix.apply(xyz.transpose()).transpose()
                x[i, :] = xyz[0, :]
                y[i, :] = xyz[1, :]
                z[i, :] = xyz[2, :]

        # translate
        x += node1[0]
        y += node1[1]
        z += node1[2]

        return x, y, z



import unittest

import numpy as np
from tnsgrt.structure import Structure


class TestStructure(unittest.TestCase):

    def test_equilibrium(self):

        nodes = np.array([[0, 0, 0], [0, 1, 0], [2, 1, 0], [2, 0, 0]]).transpose()
        members = np.array([[0, 1], [1, 2], [2, 3], [3, 0], [0, 2], [1, 3]]).transpose()
        s = Structure(nodes, members, number_of_strings=4)

        # equilibrium
        s.equilibrium()

        # no external force
        np.testing.assert_allclose(s.member_properties['lambda_'],
                                   np.hstack((np.ones((4,)), -np.ones((2,)))))
        self.assertTrue(np.abs(np.mean(
            s.get_member_properties(s.get_members_by_tag('bar'),
                                    'lambda_')) + 1) < 1e-6)

        # equilibrium
        s.equilibrium(lambda_bar=2)

        # no external force, normalize not one
        np.testing.assert_allclose(s.member_properties['lambda_'],
                                   np.hstack((2*np.ones((4,)), -2*np.ones((2,)))))
        self.assertTrue(np.abs(np.mean(
            s.get_member_properties(s.get_members_by_tag('bar'),
                                    'lambda_')) + 2) < 1e-6)

        # external force
        f = np.array([[0, 1, 0], [0, -1, 0], [0, -1, 0], [0, 1, 0]]).transpose()

        # equilibrium
        s.equilibrium(f)

        # no external force, normalize not one
        lmbda = np.array([0., 1., 0., 1., -1., -1.])
        np.testing.assert_allclose(s.member_properties['lambda_'], lmbda, atol=1e-6)

        # equilibrium
        s.equilibrium(f/2)

        # no external force, normalize not one
        lmbda = np.array([0., 1/2., 0., 1/2., -1/2., -1/2.])
        np.testing.assert_allclose(s.member_properties['lambda_'], lmbda, atol=1e-6)

        # equilibrium
        s.equilibrium(f/2, lambda_bar=2)

        # no external force, normalize not one
        lmbda = \
            np.array([0., 1/2., 0., 1/2., -1/2., -1/2.]) - \
            3/2*np.array([-1., -1., -1., -1., 1., 1])
        np.testing.assert_allclose(s.member_properties['lambda_'], lmbda, atol=1e-6)
        self.assertTrue(np.abs(np.mean(
            s.get_member_properties(s.get_members_by_tag('bar'),
                                    'lambda_')) + 2) < 1e-6)

    def test_prestress(self):

        nodes = np.array([[0, 0, 0], [0, 1, 0], [0, 2, 0]]).transpose()
        members = np.array([[0, 1], [1, 2], [0, 2]]).transpose()
        s = Structure(nodes, members, number_of_strings=2)

        # equilibrium
        s.equilibrium()

        # no external force
        np.testing.assert_allclose(s.member_properties['lambda_'],
                                   np.hstack((2*np.ones((2,)), -np.ones((1,)))))
        self.assertTrue(np.abs(np.mean(
            s.get_member_properties(s.get_members_by_tag('bar'),
                                    'lambda_')) + 1) < 1e-6)

        # equilibrium
        s.equilibrium(lambda_bar=2)

        # no external force, normalize not one
        np.testing.assert_allclose(s.member_properties['lambda_'],
                                   np.hstack((4*np.ones((2,)), -2*np.ones((1,)))))
        self.assertTrue(np.abs(np.mean(
            s.get_member_properties(s.get_members_by_tag('bar'),
                                    'lambda_')) + 2) < 1e-6)

        # external force
        f = np.array([[0, 1, 0], [0, -1, 0], [0, 0, 0]]).transpose()

        # equilibrium
        s.equilibrium(f)

        # no external force, normalize not one
        lmbda = np.array([0., 1., -1/2.])
        np.testing.assert_allclose(s.member_properties['lambda_'], lmbda, atol=1e-6)

        # equilibrium
        s.equilibrium(f/2)

        # no external force, normalize not one
        np.testing.assert_allclose(s.member_properties['lambda_'], lmbda/2, atol=1e-6)

        # equilibrium
        s.equilibrium(f, lambda_bar=2)

        # no external force, normalize not one
        lmbda = np.array([0., 1., -1/2.]) - 3/2*np.array([-2., -2., 1.])
        np.testing.assert_allclose(s.member_properties['lambda_'], lmbda, atol=1e-6)
        self.assertTrue(np.abs(np.mean(
            s.get_member_properties(
                s.get_members_by_tag('bar'), 'lambda_')) + 2) < 1e-6)



from tnsgrt.structure import Structure
from .plotter import Plotter

from vispy import scene as vispyscene


class VisPyPlotter(Plotter):
    """
    VisPy based structure plotter

    :param camera: dict with camera settings
    :param scene: dict with vispy scene settings
    """

    defaults = {
        'plot_nodes': True,
        'plot_members': True,
        'node_marker': 'o',
        'node_markersize': .05,
        'node_linewidth': 2,
        'node_linestyle': 'none',
        'node_facecolor': (1, 1, 1),
        'node_edgecolor': (1, 1, 1)
    }

    def __init__(self, camera=None, scene=None):

        scene_kwargs = {
            'keys': 'interactive',
            'size': (600, 600)
        }
        if scene:
            scene_kwargs.update(scene)

        camera_kwargs = {
            'fov': 0,
            'scale_factor': 1
        }
        if camera:
            camera_kwargs.update(camera if camera else {})

        # Create canvas and view
        self.canvas = vispyscene.SceneCanvas(**scene_kwargs)
        self.view = self.canvas.central_widget.add_view()
        self.view.camera = vispyscene.cameras.ArcballCamera(**camera_kwargs)

    def get_canvas(self):
        """
        :return: vispy canvas 
        """
        return self.canvas

    def plot(self, *s: Structure, **kwargs):

        # loop if more than one is given
        if len(s) != 1:
            for si in s:
                self.plot(si, **kwargs)
            return
        else:
            s = s[0]

        # process options
        defaults = VisPyPlotter.defaults.copy()
        defaults.update(kwargs)

        # plot nodes
        nodes = s.nodes
        if defaults['plot_nodes']:
            # Create and show visual
            vis = vispyscene.visuals.Markers(
                pos=nodes.transpose(),
                size=defaults['node_markersize'],
                antialias=0,
                face_color=defaults['node_facecolor'],
                edge_color=defaults['node_edgecolor'],
                edge_width=0,
                scaling=True,
                spherical=True,
            )
            vis.parent = self.view.scene

        # plot members
        members = s.members
        if defaults['plot_members']:
            for j in range(s.get_number_of_members()):
                if s.member_properties.loc[j, 'visible']:
                    if s.has_member_tag(j, 'string'):
                        # plot strings as lines
                        kwargs = s.get_member_properties(j, 'facecolor',
                                                         'linewidth').to_dict()
                        kwargs['color'] = kwargs['facecolor']
                        del kwargs['facecolor']
                        kwargs['width'] = kwargs['linewidth']
                        del kwargs['linewidth']
                        line = nodes[:, [members[0, j], members[1, j]]].transpose()
                        vis = vispyscene.visuals.Line(line, **kwargs)
                        vis.parent = self.view.scene
                    else:
                        # plot others as solid elements
                        kwargs = s.get_member_properties(j, 'facecolor',
                                                         'radius').to_dict()
                        kwargs['color'] = kwargs['facecolor']
                        del kwargs['facecolor']
                        line = nodes[:, [members[0, j], members[1, j]]].transpose()
                        vis = vispyscene.visuals.Tube(line, **kwargs)
                        vis.parent = self.view.scene



from typing import Optional
from typing_extensions import Literal

import numpy as np

from .structure import Structure
from .utils import rotation_2d


class Prism(Structure):
    """
    Constructs a Snelson prism with ``p`` bars

    :param p: number of bars
    :param top_radius: the top radius
    :param bottom_radius: the bottom radius
    :param height: the radius of the prism
    :param alpha: the twist angle
    :param calculate_equilibrium: if ``True`` calculates equilibrium
    :param equilibrium_method: choice of method for computing equilibrium
    :param \\**kwargs: See below

    :Keyword Arguments:
        * **bar** (``bool=True``) --
          if ``True`` add *bars*
        * **bottom** (``bool=True``) --
          if ``True`` add *bottom strings*
        * **top** (``bool=True``) --
          if ``True`` add *top strings*
        * **vertical** (``bool=True``) --
          if ``True`` add *vertical strings*
        * **diagonal** (``bool=False``) --
          if ``True`` add *diagonal strings*

    **Notes:**

    1. Unloaded equilibrium is possible only if

       .. math::
          \\frac{\\pi}{2} - \\frac{\\pi}{p} \\leq \\alpha \\leq \\frac{\\pi}{2}

       when both *vertical* and *diagonal* strings are present
    2. If *diagonal* strings are not present, then unloaded equilibrium is possible
       only if

       .. math::
          \\alpha = \\frac{\\pi}{2} - \\frac{\\pi}{p}

       This is the default prism configuration
    3. If *vertical* strings are not present, then unloaded equilibrium is possible
       only if

       .. math::
          \\alpha = \\frac{\\pi}{2}
    4. Additional keyword arguments are passed to :class:`tnsgrt.structure.Structure`
    """

    def __init__(self, p: int = 3,
                 top_radius: float = 1, bottom_radius: float = 1,
                 height: float = 1, alpha: Optional[float] = None,
                 calculate_equilibrium=True,
                 equilibrium_method=Literal['analytic', 'numeric'],
                 **kwargs):

        # proper size
        assert p >= 3, 'p must be greater or equal to 3'

        # twist angle
        if alpha is None:
            alpha = np.pi/2 - np.pi/p

        # other options
        options = {
            'bottom': True,
            'top': True,
            'vertical': True,
            'diagonal': False,
            'bar': True
        }
        options.update(kwargs)

        # clean up kwargs
        kwargs = {k: v for k, v in kwargs.items()
                  if k not in ['bottom', 'top', 'vertical', 'diagonal', 'bar']}

        # valid equilibrium?
        if calculate_equilibrium:
            assert 'bottom' in options, \
                'equilibrium is not possible without bottom strings'
            assert 'top' in options, 'equilibrium is not possible without top strings'
            assert 'bar' in options, 'equilibrium is not possible without bars'
            if not options['diagonal']:
                assert np.abs(np.pi/2 - np.pi/p - alpha) < 1e-8, \
                    'equilibrium is not possible with given twist angle'
            if not options['vertical']:
                assert np.abs(np.pi/2 - alpha) < 1e-8, \
                    'equilibrium is not possible with given twist angle'
            if options['vertical'] and options['diagonal']:
                assert np.pi/2 - np.pi/p <= alpha <= np.pi/2,\
                    'equilibrium is not possible with given twist angle'

        # base angle
        beta = 2 * np.pi/p

        # rotation matrices
        rotation_beta = rotation_2d(beta)
        rotation_alpha = rotation_2d(alpha)

        # nodes
        nodes = np.zeros((3, 2 * p))
        # bottom nodes
        nodes[2, :p] = -height / 2
        nodes[:2, 0] = np.array([bottom_radius, 0])
        for i in range(1, p):
            nodes[:2, i] = rotation_beta @ nodes[:2, i-1]
        # top nodes
        nodes[2, p:2 * p] = height / 2
        nodes[:2, p] = np.array([top_radius, 0])
        nodes[:2, p] = rotation_alpha @ nodes[:2, p]
        for i in range(1, p):
            nodes[:2, p + i] = rotation_beta @ nodes[:2, p + i - 1]

        # strings
        members = np.zeros((2, 0))
        number_of_strings = 0
        string_tags = {}
        # bottom strings
        bottom_strings = np.vstack((np.arange(p), np.mod(1 + np.arange(p), p)))
        if options['bottom']:
            members = np.hstack((members, bottom_strings))
            string_tags['bottom'] = number_of_strings + np.arange(p)
            number_of_strings += p
        # top strings
        top_strings = np.vstack((p + np.arange(p), p + np.mod(1 + np.arange(p), p)))
        if options['top']:
            members = np.hstack((members, top_strings))
            string_tags['top'] = number_of_strings + np.arange(p)
            number_of_strings += p
        # vertical strings
        vertical_strings = np.vstack((np.arange(p), p + np.arange(p)))
        if options['vertical']:
            members = np.hstack((members, vertical_strings))
            string_tags['vertical'] = number_of_strings + np.arange(p)
            number_of_strings += p
        # diagonal strings
        diagonal_strings = np.vstack((np.arange(p), p + np.mod(np.arange(p) - 1, p)))
        if options['diagonal']:
            members = np.hstack((members, diagonal_strings))
            string_tags['diagonal'] = number_of_strings + np.arange(p)
            number_of_strings += p

        # bars
        bars = np.vstack((np.arange(p), p + np.mod(1 + np.arange(p), p)))
        if options['bar']:
            members = np.hstack((members, bars))

        # call super
        super().__init__(nodes=nodes, members=members,
                         number_of_strings=number_of_strings,
                         member_tags=string_tags, **kwargs)

        if calculate_equilibrium:
            if equilibrium_method == 'analytic':
                # top/bottom ratio
                rho = top_radius / bottom_radius
                # force coefficients
                lambda_ = np.array([
                    # bottom
                    rho * np.cos(np.pi / p) / np.cos(alpha - np.pi / p),
                    # top
                    (1 / rho) * np.cos(np.pi / p) / np.cos(alpha - np.pi / p),
                    # vertical
                    2 * np.cos(alpha) * np.cos(np.pi / p) / np.cos(alpha - np.pi / p),
                    # diagonal
                    -np.cos(alpha + np.pi / p) / np.cos(alpha - np.pi / p),
                    # bars
                    -1
                ]).reshape((1, 5))
                # select appropriate string sets
                index = [True, True, options['vertical'], options['diagonal'], True]
                # construct coefficients
                self.member_properties['lambda_'] = \
                    (np.ones((p, 1)) @ lambda_[:, index]).flatten(order='F')
            else:
                # solve for equilibrium
                self.equilibrium(equalities=[np.arange(number_of_strings,
                                                       members.shape[1])])

            # update mass, volume, stiffness and rest length
            self.update_member_properties()
